# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module defines the Page for the physical properties of the fluid.
These properties can be reference value or initial value

This module contens the following classes and function:
- FluidCaracteristicsModel
- FluidCaracteristicsModelTestCase
"""

#-------------------------------------------------------------------------------
# Library modules import
#-------------------------------------------------------------------------------

import os, sys, unittest

#-------------------------------------------------------------------------------
# Application modules import
#-------------------------------------------------------------------------------

from code_saturne.model.Common import *
from code_saturne.model.XMLvariables import Variables, Model
from code_saturne.model.XMLmodel import XMLmodel, ModelTest

from code_saturne.model.ThermalScalarModel import ThermalScalarModel
from code_saturne.model.DefineUserScalarsModel import DefineUserScalarsModel
from code_saturne.model.NotebookModel import NotebookModel

#-------------------------------------------------------------------------------
# EOS
#-------------------------------------------------------------------------------

EOS = 0
try:
   import eosAva
   EOS = 1
except Exception:
   pass

# Perfect gases to exclude from EOS choices

eos_excl = ["Argon", "Nitrogen", "Hydrogen", "Oxygen", "Helium", "Air"]

#-------------------------------------------------------------------------------
# Coolprop
#-------------------------------------------------------------------------------

from code_saturne import cs_config

coolprop_fluids = []
coolprop_warn = False

if cs_config.config().libs['coolprop'].have != "no" and not coolprop_fluids:

   try:
      import sys
      sys.path.insert(0, cs_config.config().libs['coolprop'].flags['pythonpath'])
      import CoolProp
      sys.path.pop(0)
      self.coolprop_fluids = []
      for f in CoolProp.__fluids__:
         coolprop_fluids.append(f)
      coolprop_fluids.sort()

   except Exception:  # CoolProp might be available but not its Python bindings

      if cs_config.config().libs['coolprop'].have != "gui_only":
         """
         import traceback
         exc_info = sys.exc_info()
         bt = traceback.format_exception(*exc_info)
         for l in bt:
            print(l)
         del exc_info
         print("Warning: CoolProp Python bindings not available or usable")
         print("         list of fluids based on CoolProp 5.1.1")
         """
         pass
      else:
         coolprop_warn = True

      coolprop_fluids = ['1-Butene', 'Acetone', 'Air', 'Ammonia', 'Argon',
                         'Benzene', 'CarbonDioxide', 'CarbonMonoxide',
                         'CarbonylSulfide', 'CycloHexane', 'CycloPropane',
                         'Cyclopentane', 'D4', 'D5', 'D6', 'Deuterium',
                         'DimethylCarbonate', 'DimethylEther', 'Ethane',
                         'Ethanol', 'EthylBenzene', 'Ethylene', 'Fluorine',
                         'HFE143m', 'HeavyWater', 'Helium', 'Hydrogen',
                         'HydrogenSulfide', 'IsoButane', 'IsoButene',
                         'Isohexane', 'Isopentane', 'Krypton', 'MD2M', 'MD3M',
                         'MD4M', 'MDM', 'MM', 'Methane', 'Methanol',
                         'MethylLinoleate', 'MethylLinolenate', 'MethylOleate',
                         'MethylPalmitate', 'MethylStearate', 'Neon',
                         'Neopentane', 'Nitrogen', 'NitrousOxide', 'Novec649',
                         'OrthoDeuterium', 'OrthoHydrogen', 'Oxygen',
                         'ParaDeuterium', 'ParaHydrogen', 'Propylene',
                         'Propyne', 'R11', 'R113', 'R114', 'R115', 'R116',
                         'R12', 'R123', 'R1233zd(E)', 'R1234yf', 'R1234ze(E)',
                         'R1234ze(Z)', 'R124', 'R125', 'R13', 'R134a', 'R13I1',
                         'R14', 'R141b', 'R142b', 'R143a', 'R152A', 'R161',
                         'R21', 'R218', 'R22', 'R227EA', 'R23', 'R236EA',
                         'R236FA', 'R245fa', 'R32', 'R365MFC', 'R404A',
                         'R407C', 'R41', 'R410A', 'R507A', 'RC318', 'SES36',
                         'SulfurDioxide', 'SulfurHexafluoride', 'Toluene',
                         'Water', 'Xenon', 'cis-2-Butene', 'm-Xylene',
                         'n-Butane', 'n-Decane', 'n-Dodecane', 'n-Heptane',
                         'n-Hexane', 'n-Nonane', 'n-Octane', 'n-Pentane',
                         'n-Propane', 'n-Undecane', 'o-Xylene', 'p-Xylene',
                         'trans-2-Butene']

#-------------------------------------------------------------------------------
# Model class
#-------------------------------------------------------------------------------

class FluidCharacteristicsModel(Variables, Model):
    """
    Class to manipulate Molecular Properties in xml file.
    """
    def __init__(self, case):
        """FluidCharacteristicsModel Constuctor."""
        self.case = case
        self.node_models = self.case.xmlGetNode('thermophysical_models')
        self.node_prop   = self.case.xmlGetNode('physical_properties')
        self.node_fluid  = self.node_prop.xmlInitNode('fluid_properties')
        self.node_comp   = self.node_models.xmlInitNode('compressible_model', 'model')
        self.node_gas    = self.node_models.xmlInitNode('gas_combustion',     'model')
        self.node_coal   = self.node_models.xmlInitNode('solid_fuels',        'model')

        # Info on available libraries

        self.mask_builtin   = 1 << 0
        self.mask_CoolProp  = 1 << 1
        self.mask_EOS       = 1 << 2
        self.mask_freesteam = 1 << 3

        self.tables = 0

        from code_saturne import cs_config
        cfg = cs_config.config()

        self.lib_properties = {}
        self.lib_properties['user_material'] = self.mask_builtin

        if EOS == 1:
            self.tables += self.mask_EOS
            self.ava = eosAva.EosAvailable()
            # suppress perfect gas
            fls = self.ava.whichFluids()
            for fli in fls:
                if fli in eos_excl:
                    continue
                if fli not in self.lib_properties.keys():
                    self.lib_properties[fli] = self.mask_EOS
                else:
                    self.lib_properties[fli] += self.mask_EOS

        if cfg.libs['freesteam'].have != "no":
            self.tables += self.mask_freesteam
            fli = 'Water'
            if fli not in self.lib_properties.keys():
                self.lib_properties[fli] = self.mask_freesteam
            else:
                self.lib_properties[fli] += self.mask_freesteam

        if coolprop_fluids:
            if cfg.libs['coolprop'].have != "no":
                self.tables += self.mask_CoolProp
        for fli in coolprop_fluids:
            if fli not in self.lib_properties.keys():
                self.lib_properties[fli] = self.mask_CoolProp
            else:
                self.lib_properties[fli] += self.mask_CoolProp

        # Base model needs density and molecular viscosity

        self.lst = [('density', 'Rho'),('molecular_viscosity', 'Mu')]
        self.node_density   = self.setNewFluidProperty(self.node_fluid, \
                                                       'density')
        self.node_viscosity = self.setNewFluidProperty(self.node_fluid, \
                                                       'molecular_viscosity')
        self.node_lst = [self.node_density, self.node_viscosity]

        self.node_heat = None
        self.node_cond = None
        self.node_vol_visc = None
        self.node_dyn = None

        # Get thermal scalar and model

        thm = ThermalScalarModel(self.case)
        tsn = thm.getThermalScalarName()
        self.tsm = thm.getThermalScalarModel()

        # If thermal model enabled, add thermal conductivity and specific heat

        if self.tsm != "off":
            self.lst.extend([('specific_heat', 'Cp'), \
                             ('thermal_conductivity', 'Al')])
            self.node_heat = self.setNewFluidProperty(self.node_fluid, \
                                                      'specific_heat')
            self.node_cond = self.setNewFluidProperty(self.node_fluid, \
                                                      'thermal_conductivity')
            self.node_lst.extend([self.node_heat, self.node_cond])

        # Define volume viscosity for compressible model

        if self.node_comp['model'] not in [None, "off"]:
            self.lst.append(('volume_viscosity', 'Viscv0'))
            self.node_vol_visc  = self.setNewFluidProperty(self.node_fluid, \
                                                           'volume_viscosity')
            self.node_lst.append(self.node_vol_visc)

        # Define dynamic diffusion for reactive flow
        elif self.node_coal['model'] not in [None, "off"] \
             or self.node_gas['model'] not in [None, "off"]:
            self.lst.append(('dynamic_diffusion', 'Diftl0'))
            self.node_dyn = self.setNewFluidProperty(self.node_fluid, \
                                                     'dynamic_diffusion')
            self.node_lst.append(self.node_dyn)

        # Build scalars list

        self.list_scalars = []

        if self.tsm == "temperature_celsius":
            self.list_scalars.append((tsn, self.tr("Thermal scalar: temperature (\xB0 C)")))
        elif self.tsm == "temperature_kelvin":
            self.list_scalars.append((tsn, self.tr("Thermal scalar: temperature (K)")))
        elif self.tsm != "off":
            self.list_scalars.append((tsn, self.tr("Thermal scalar")))

        self.m_sca = DefineUserScalarsModel(self.case)
        for s in self.m_sca.getUserScalarNameList():
            self.list_scalars.append((s, self.tr("Additional scalar")))

        # Notebook

        self.notebook = NotebookModel(self.case)


    def __nodeFromTag(self, name):
        """
        Private method : return node with attibute name 'name'
        """
        for node in self.node_lst:
            if node['name'] == name:
                return node


    def getLibPropertiesDict(self):
       """
       Return dictionnary with available properties and matching flags
       """
       return self.lib_properties


    def getLibPropertyMethods(self, material):
        """
        Return list of methods for a given library and material.
        Values are returned as a tuple: (name, available)
        """
        methods = []

        if material not in self.lib_properties.keys():
            methods = [('unknown', False)]
            return methods

        material_flags = self.lib_properties[material]

        if material_flags & self.mask_builtin:
            methods.append(('user_properties', True))

        if material_flags & self.mask_EOS:
            if material not in eos_excl:
                self.ava.setMethods(material)
                fls = self.ava.whichMethods()
                for fli in fls:
                    if fli != 'PerfectGas':
                        methods.append((fli, True))

        if material_flags & self.mask_freesteam:
            avail = self.tables & self.mask_freesteam != 0
            methods.append(("freesteam", avail))

        if material_flags & self.mask_CoolProp:
            avail = self.tables & self.mask_CoolProp != 0
            avail = False
            methods.append(("CoolProp", avail))

        return methods


    def defaultFluidCharacteristicsValues(self):
        """
        Return in a dictionnary which contains default values.
        (also called by CoalCombustionModel)
        """
        default = {}

        # Particular default values init. taking into account chosen model
        mdl_atmo, mdl_joule, mdl_thermal, mdl_gas, mdl_coal, mdl_comp, mdl_hgn=\
            self.getThermoPhysicalModel()

        default['reference_pressure']    = 1.01325e+5

        if self.tsm == "temperature_celsius":
           default['reference_temperature'] = 20.
        else:
           default['reference_temperature'] = 293.15

        default['reference_fuel_temperature']      = 436.
        default['reference_oxydant_temperature']   = 353.
        # molar mass for dry air
        default['reference_molar_mass'] = 28.966e-3

        # Initial values for properties: 20 Celsius degrees air at atmospheric
        # pressure except for VoF

        default['density']                = 1.17862
        default['density_0']              = 1000.
        default['density_1']              = 1.
        default['molecular_viscosity']    = 1.83e-05
        default['molecular_viscosity_0']  = 1.e-03
        default['molecular_viscosity_1']  = 1.e-05
        if mdl_hgn != "off":
           default['density']             = default['density_1']
           default['molecular_viscosity'] = default['molecular_viscosity_1']
        default['specific_heat']          = 1017.24
        default['thermal_conductivity']   = 0.02495
        default['dynamic_diffusion']      = 0.01
        default['volume_viscosity']       = 0.
        default['material']               = "user_material"
        default['method']                 = "user_properties"
        default['reference']              = None

        return default

    @Variables.undoLocal
    def setPressure(self, value):
        """
        Set value of reference pressure into xml file.
        """
        self.isGreaterOrEqual(value, 0.0)
        self.node_fluid.xmlSetData('reference_pressure', value)


    @Variables.noUndo
    def getPressure(self):
        """
        Return the value of reference pressure.
        """
        value = self.node_fluid.xmlGetDouble('reference_pressure')
        if value == None:
           p_str = 'reference_pressure'
           value = self.defaultFluidCharacteristicsValues()[p_str]
           self.setPressure(value)

        return value


    @Variables.undoLocal
    def setTemperature(self, value):
        """
        Set reference temperature.
        """
        self.isGreater(value, 0.0)
        self.node_fluid.xmlSetData('reference_temperature', value)


    @Variables.noUndo
    def getTemperature(self):
        """
        Get reference temperature.
        """
        value = self.node_fluid.xmlGetDouble('reference_temperature')
        if value == None:
           t_str = 'reference_temperature'
           value = self.defaultFluidCharacteristicsValues()[t_str]
           self.setTemperature(value)
        return value


    @Variables.undoLocal
    def setTempOxydant(self, value):
        """
        Set reference temperature for Oxydant.
        """
        self.isGreater(value, 0.0)
        self.node_fluid.xmlSetData('reference_oxydant_temperature', value)


    @Variables.noUndo
    def getTempOxydant(self):
        """
        Get reference temperaturefor Oxydant.
        """
        value = self.node_fluid.xmlGetDouble('reference_oxydant_temperature')
        if value == None:
           ot_str = 'reference_oxydant_temperature'
           value = self.defaultFluidCharacteristicsValues()[ot_str]
           self.setTempOxydant(value)
        return value


    @Variables.undoLocal
    def setTempFuel(self, value):
        """
        Set reference temperature.
        """
        self.isGreater(value, 0.0)
        self.node_fluid.xmlSetData('reference_fuel_temperature', value)


    @Variables.noUndo
    def getTempFuel(self):
        """
        Get reference temperature.
        """
        value = self.node_fluid.xmlGetDouble('reference_fuel_temperature')
        if value == None:
           ft_str = 'reference_fuel_temperature'
           value = self.defaultFluidCharacteristicsValues()[ft_str]
           self.setTempFuel(value)
        return value


    @Variables.undoLocal
    def setMassemol(self, value):
        """
        Set reference molar mass.
        """
        self.isGreater(value, 0.0)
        self.node_fluid.xmlSetData('reference_molar_mass', value)


    @Variables.noUndo
    def getMassemol(self):
        """
        Get reference molar mass.
        """
        value = self.node_fluid.xmlGetDouble('reference_molar_mass')
        if value == None:
           mm_str = 'reference_molar_mass'
           value = self.defaultFluidCharacteristicsValues()[mm_str]
           self.setMassemol(value)
        return value


    @Variables.noUndo
    def getThermoPhysicalModel(self):
        """
        Return values of attribute "model" of all thermophysical model nodes.
        (also called by NumericalParamGlobalView and TimeStepView)
        """
        d = {}

        for mdl in ['atmospheric_flows', 'compressible_model',
                    'gas_combustion', 'joule_effect', 'solid_fuels',
                    'thermal_scalar', 'hgn_model']:
            d[mdl] = 'off'

            node = self.node_models.xmlGetNode(mdl, 'model')

            if node:
                if node['model'] == "":
                    node['model'] = "off"
                if node['model'] != 'off':
                    d[mdl] = node['model']

        return d['atmospheric_flows'], \
               d['joule_effect'],      \
               d['thermal_scalar'],    \
               d['gas_combustion'],    \
               d['solid_fuels'],       \
               d['compressible_model'],\
               d['hgn_model']


    @Variables.noUndo
    def getMaterials(self):
        """
        get the nature of materials
        """
        nodem = self.node_fluid.xmlGetNode('material')
        if nodem == None:
            material = self.defaultFluidCharacteristicsValues()['material']
            self.setMaterials(material)
            nodem = self.node_fluid.xmlGetNode('material')
        material = nodem['choice']
        return material


    @Variables.undoLocal
    def setMaterials(self, material):
        """
        set the nature of materials
        """
        childNode = self.node_fluid.xmlInitChildNode('material')
        m = childNode.xmlGetNode('material')
        oldMaterial = None
        if m != None:
            oldMaterial = m['choice']
        childNode.xmlSetAttribute(choice = material)
        self.updateMethod(oldMaterial)

        if material == "user_material":
            for node in self.node_fluid.xmlGetChildNodeList('property'):
                if node['choice'] == "thermal_law":
                    node['choice'] = "user_law"


    @Variables.noUndo
    def getMethod(self):
        """
        get the method used to compute properties
        """
        nodem = self.node_fluid.xmlGetNode('method')
        if nodem == None:
            method = self.updateMethod("")
            nodem = self.node_fluid.xmlGetNode('method')
        method = nodem['choice']
        return method


    @Variables.undoLocal
    def setMethod(self, method):
        """
        update reference value for EOS
        """
        childNode = self.node_fluid.xmlInitChildNode('method')
        childNode.xmlSetAttribute(choice = method)
        self.getReference()

        # suppress phase choice (not used)
        nodem = self.node_fluid.xmlGetNode('phas')
        if nodem:
            nodem.xmlRemoveNode()

        # suppress reference choice if not EOS
        if method in ("user_properties", "freesteam", "CoolProp"):
            nodem = childNode.xmlGetNode('reference')
            if nodem:
                nodem.xmlRemoveNode()


    @Variables.undoGlobal
    def updateMethod(self, oldMaterial):
        """
        update reference value for EOS
        """
        material = self.getMaterials()
        if oldMaterial == material:
            return

        old_method = self.getMethod

        # If fluid not known, do no change anything (error may be caught at
        # computation time, or the GUI may simply be incomplete)
        if material not in self.lib_properties.keys():
            return

        material_flags = self.lib_properties[material]
        methods = []

        if material_flags & self.mask_builtin:
            methods.append('user_properties')

        if material_flags & self.mask_EOS:
            self.ava.setMethods(material)
            fls = self.ava.whichMethods()
            for fli in fls:
                methods.append(fli)

        if material_flags & self.mask_CoolProp:
            methods.append("CoolProp")

        if material_flags & self.mask_freesteam:
            methods.append("freesteam")

        if old_method not in methods and methods:
            methods.append("unknown")
            self.setMethod(methods[0])

        reference = self.getReference() # to force update


    @Variables.noUndo
    def getAvailReferences(self, material, method):
        """
        return available reference value for EOS
        """
        if method in ("user_properties", "freesteam", "CoolProp"):
            return None

        references = []

        if EOS:
            self.ava = eosAva.EosAvailable()
            self.ava.setMethods(material)
            self.ava.setReferences(material, self.getMethod())
            references = self.ava.whichReferences()
        return references


    @Variables.noUndo
    def getReference(self):
        """
        return reference value for EOS
        """
        reference = ""
        material = self.getMaterials()
        method = self.getMethod()
        if method in ("user_properties", "freesteam", "CoolProp"):
            return None

        nodem = self.node_fluid.xmlGetChildNode('method')
        reference = nodem.xmlGetString('reference')

        if not reference and self.lib_properties[material] & self.mask_EOS and EOS:
            self.ava = eosAva.EosAvailable()
            self.ava.setMethods(material)
            self.ava.setReferences(material, self.getMethod())
            ref = self.ava.whichReferences()
            if ref:
                reference = ref[0]

            # update XML
            self.setReference('reference')

        return reference


    @Variables.noUndo
    def setReference(self, reference):
        """
        set reference value for EOS
        """
        nodem = self.node_fluid.xmlInitChildNode('method')
        if not reference:
            if nodem.xmlGetNode('reference'):
                node.xmlRemoveChild('reference')
        else:
            nodem.xmlSetData('reference', reference)


    @Variables.noUndo
    def getInitialValue(self, tag):
        """
        Return initial value of the markup tag : 'density', or
        'molecular_viscosity', or 'specific_heat'or 'thermal_conductivity'
        """
        self.isInList(tag, ('density', 'molecular_viscosity',
                            'specific_heat', 'thermal_conductivity',
                            'volume_viscosity', 'dynamic_diffusion'))
        node = self.node_fluid.xmlGetNode('property', name=tag)
        pp = node.xmlGetDouble('initial_value')
        if pp == None:
            pp = self.defaultFluidCharacteristicsValues()[tag]
            self.setInitialValue(tag, pp)
        return pp


    @Variables.noUndo
    def getValue(self, f_id, tag):
        """
        Return value of the markup tag : 'density', or
        'molecular_viscosity' for fluid of id f_id
        """
        self.isInList(tag, ('density', 'molecular_viscosity'))
        node = self.node_fluid.xmlGetNode('property', name=tag)
        pp = node.xmlGetDouble('value' + '_' + str(f_id))
        if pp == None:
            pp = self.defaultFluidCharacteristicsValues()[tag+'_'+str(f_id)]
            self.setValue(f_id, tag, pp)
        return pp


    @Variables.undoLocal
    def setInitialValue(self, tag, val):
        """
        Set initial value for the markup tag : 'density', or
        'molecular_viscosity', or 'specific_heat'or 'thermal_conductivity'
        """
        self.isInList(tag, ('density', 'molecular_viscosity',
                            'specific_heat', 'thermal_conductivity',
                            'volume_viscosity', 'dynamic_diffusion'))
        if tag != 'volume_viscosity':
            self.isGreater(val, 0.)
        else:
            self.isPositiveFloat(val)
        node = self.node_fluid.xmlGetNode('property', name=tag)
        node.xmlSetData('initial_value', val)


    @Variables.undoLocal
    def setValue(self, f_id, tag, val):
        """
        Set initial value for the markup tag : 'density', or
        'molecular_viscosity'
        """
        self.isInList(tag, ('density', 'molecular_viscosity'))
        node = self.node_fluid.xmlGetNode('property', name=tag)
        node.xmlSetData('value'+'_'+str(f_id), val)


    @Variables.noUndo
    def getInitialValueDensity(self):
        """Return initial value of density"""
        return self.getInitialValue('density')


    @Variables.noUndo
    def getVofValueDensity(self, f_id):
        """Return value of density of fluid of id f_id"""
        return self.getValue(f_id, 'density')


    @Variables.undoLocal
    def setInitialValueDensity(self, val):
        """Put initial value for density"""
        self.setInitialValue('density', val)


    @Variables.undoLocal
    def setVofValueDensity(self, f_id, val):
        """Set value for density of fluid of id f_id"""
        self.setValue(f_id, 'density', val)


    @Variables.noUndo
    def getInitialValueViscosity(self):
        """Return initial value of viscosity"""
        return self.getInitialValue('molecular_viscosity')


    @Variables.noUndo
    def getVofValueViscosity(self, f_id):
        """Return value of viscosity of fluid of id f_id"""
        return self.getValue(f_id, 'molecular_viscosity')


    @Variables.undoLocal
    def setInitialValueViscosity(self, val):
        """Put initial value for viscosity"""
        self.setInitialValue('molecular_viscosity', val)


    @Variables.undoLocal
    def setVofValueViscosity(self, f_id, val):
        """Set value for viscosity of fluid of id f_id"""
        self.setValue(f_id, 'molecular_viscosity', val)


    @Variables.noUndo
    def getInitialValueVolumeViscosity(self):
        """Return initial value of volume viscosity"""
        return self.getInitialValue('volume_viscosity')


    @Variables.undoLocal
    def setInitialValueVolumeViscosity(self, val):
        """Put initial value for volume viscosity"""
        self.setInitialValue('volume_viscosity', val)


    @Variables.noUndo
    def getInitialValueHeat(self):
        """Return initial value of specific heat"""
        return self.getInitialValue('specific_heat')


    @Variables.undoLocal
    def setInitialValueHeat(self, val):
        """Put initial value for specific heat"""
        self.setInitialValue('specific_heat', val)


    @Variables.noUndo
    def getInitialValueCond(self):
        """Return initial value of conductivity"""
        return self.getInitialValue('thermal_conductivity')


    @Variables.undoLocal
    def setInitialValueCond(self, val):
        """Put initial value for conductivity"""
        self.setInitialValue('thermal_conductivity', val)


    @Variables.noUndo
    def getInitialValueDyn(self):
        """Return initial value of conductivity"""
        return self.getInitialValue('dynamic_diffusion')


    @Variables.undoLocal
    def setInitialValueDyn(self, val):
        """Put initial value for conductivity"""
        self.setInitialValue('dynamic_diffusion', val)


    @Variables.noUndo
    def getFormulaTry(self, tag, zone="1"):
        """
        Return a formula for I{tag} 'density', 'molecular_viscosity',
        'specific_heat' or 'thermal_conductivity'
        only if actually defined, None otherwise.
        """
        self.isInList(tag, ('density', 'dynamic_diffusion',
                            'molecular_viscosity',
                            'specific_heat', 'thermal_conductivity',
                            'volume_viscosity'))
        node = self.node_fluid.xmlGetNode('property', name=tag)
        if zone != "1":
            if node.xmlGetChildNode("zone", zone_id=zone):
                node = node.xmlGetChildNode("zone", zone_id=zone)
            else:
                node = node.xmlInitChildNode("zone", zone_id=zone)
        formula = node.xmlGetChildString('formula')
        return formula


    @Variables.noUndo
    def getFormula(self, tag, zone="1"):
        """
        Return a formula for I{tag} 'density', 'molecular_viscosity',
        'specific_heat' or 'thermal_conductivity'
        """
        formula = self.getFormulaTry(tag, zone)
        if not formula:
            formula = self.getDefaultFormula(tag)
            self.setFormula(tag, formula, zone)
        return formula


    @Variables.noUndo
    def getDefaultFormula(self, tag):
        """
        Return default formula
        """
        self.isInList(tag, ('density', 'molecular_viscosity',
                            'specific_heat', 'thermal_conductivity',
                            'volume_viscosity'))

        d_value = {'density': 'rho0',
                   'molecular_viscosity': 'mu0',
                   'specific_heat': 'cp0',
                   'thermal_conductivity': 'lambda0',
                   'volume_viscosity': 'viscv0'}

        formula = tag + " = " + d_value[tag] + ";"

        return formula


    @Variables.undoLocal
    def setFormula(self, tag, formula, zone="1"):
        """
        Gives a formula for 'density', 'molecular_viscosity',
        'specific_heat'or 'thermal_conductivity'
        """
        self.isInList(tag, ('density', 'molecular_viscosity',
                            'specific_heat', 'thermal_conductivity',
                            'volume_viscosity'))
        node = self.node_fluid.xmlGetNode('property', name=tag)
        if zone != "1":
            if node.xmlGetChildNode("zone", zone_id=zone):
                node = node.xmlGetChildNode("zone", zone_id=zone)
            else:
                node = node.xmlInitChildNode("zone", zone_id=zone)
        node.xmlSetData('formula', formula)


    @Variables.noUndo
    def getPropertyMode(self, tag):
        """Return choice of node I{tag}. Choice is constant or variable"""
        self.isInList(tag, ('density', 'molecular_viscosity',
                            'specific_heat', 'thermal_conductivity',
                            'volume_viscosity', 'dynamic_diffusion'))
        node = self.__nodeFromTag(tag)
        c = node['choice']
        self.isInList(c, ('constant', 'thermal_law', 'user_law', 'predefined_law'))
        return c


    @Variables.undoGlobal
    def setPropertyMode(self, tag, choice):
        """Put choice in xml file's node I{tag}"""
        self.isInList(tag, ('density', 'molecular_viscosity',
                            'specific_heat', 'thermal_conductivity',
                            'volume_viscosity', 'dynamic_diffusion'))
        self.isInList(choice, ('constant', 'thermal_law',
                               'user_law', 'predefined_law'))

        node = self.__nodeFromTag(tag)
        node['choice'] = choice

        if choice == 'constant':

            if tag == 'density':
                from code_saturne.model.TimeStepModel import TimeStepModel
                TimeStepModel(self.case).RemoveThermalTimeStepNode()
                del TimeStepModel
        else:
            if node.xmlGetNode('listing_printing'):
                node.xmlRemoveChild('listing_printing')
            if node.xmlGetNode('postprocessing_recording'):
                node.xmlRemoveChild('postprocessing_recording')

    # MEG Generation related functions
    def getFormulaComponents(self, tag, scalar=None, zone="1"):
        """
        Get the formula components for a given tag
        """

        if tag == 'density':
            return self.getFormulaRhoComponents(zone)

        elif tag == 'molecular_viscosity':
            return self.getFormulaMuComponents(zone)

        elif tag == 'specific_heat':
            return self.getFormulaCpComponents(zone)

        elif tag == 'thermal_conductivity':
            return self.getFormulaAlComponents(zone)

        elif tag == 'volume_viscosity':
            return self.getFormulaViscv0Components(zone)

        elif tag == 'scalar_diffusivity' and scalar != None:
            return self.getFormulaDiffComponents(scalar, zone)

        else:
            msg = 'Formula is not available for field %s in MEG' % tag
            raise Exception(msg)


    def getFormulaRhoComponents(self, zone="1"):
        """
        User formula for density
        """

        exp = self.getFormula('density', zone)
        req = [('density', 'Density')]

        symbols = []
        for s in self.list_scalars:
           symbols.append(s)

        rho0_value = self.getInitialValueDensity()
        symbols.append(('rho0', 'Density (reference value) = ' + str(rho0_value)))

        ref_pressure = self.getPressure()
        symbols.append(('p0', 'Reference pressure = ' + str(ref_pressure)))

        symbols.append(('volume', 'Zone volume'))
        symbols.append(('fluid_volume', 'Zone fluid volume (different if porous)'))

        for (nme, val) in self.notebook.getNotebookList():
            symbols.append((nme, 'value (notebook) = ' + str(val)))

        return exp, req, self.list_scalars, symbols;


    def getFormulaMuComponents(self, zone="1"):
        """
        User formula for molecular viscosity
        """

        exp = self.getFormula('molecular_viscosity', zone)
        req = [('molecular_viscosity', 'Molecular Viscosity')]

        symbols = []

        for s in self.list_scalars:
           symbols.append(s)
        mu0_value = self.getInitialValueViscosity()
        symbols.append(('mu0', 'Viscosity (reference value) = ' + str(mu0_value)))

        rho0_value = self.getInitialValueDensity()
        symbols.append(('rho0', 'Density (reference value) = ' + str(rho0_value)))

        ref_pressure = self.getPressure()
        symbols.append(('p0', 'Reference pressure = ' + str(ref_pressure)))

        symbols.append(('volume', 'Zone volume'))
        symbols.append(('fluid_volume', 'Zone fluid volume (different if porous)'))

        from code_saturne.model.CompressibleModel import CompressibleModel
        if CompressibleModel(self.case).getCompressibleModel() == 'on':
            symbols.append(('T', 'Temperature'))
            ref_temperature = self.getTemperature()
            symbols.append(('t0', 'Reference temperature = '+str(ref_temperature)+' K'))

        for (nme, val) in self.notebook.getNotebookList():
            symbols.append((nme, 'value (notebook) = ' + str(val)))

        return exp, req, self.list_scalars, symbols;


    def getFormulaCpComponents(self, zone="1"):
        """
        User formula for specific heat
        """
        exp = self.getFormula('specific_heat', zone)
        req = [('specific_heat', 'Specific heat')]

        symbols = []
        for s in self.list_scalars:
           symbols.append(s)

        cp0_value = self.getInitialValueHeat()
        symbols.append(('cp0', 'Specific heat (reference value) = ' + str(cp0_value)))

        ref_pressure = self.getPressure()
        symbols.append(('p0', 'Reference pressure = ' + str(ref_pressure)))

        symbols.append(('volume', 'Zone volume'))
        symbols.append(('fluid_volume', 'Zone fluid volume (different if porous)'))

        for (nme, val) in self.notebook.getNotebookList():
            symbols.append((nme, 'value (notebook) = ' + str(val)))

        return exp, req, self.list_scalars, symbols;


    def getFormulaViscv0Components(self, zone="1"):
        """
        User formula for volumic viscosity
        """
        exp = self.getFormula('volume_viscosity', zone)
        req = [('volume_viscosity', 'Volume viscosity')]

        symbols = []
        for s in self.list_scalars:
           symbols.append(s)

        viscv0_value = self.getInitialValueVolumeViscosity()
        ref_pressure = self.getPressure()
        ref_temperature = self.getTemperature()
        symbols.append(('viscv0', 'Volume viscosity (J/kg/K) = '+str(viscv0_value)))
        symbols.append(('p0', 'Reference pressure (Pa) = '+str(ref_pressure)))
        symbols.append(('t0', 'Reference temperature (K) = '+str(ref_temperature)))
        symbols.append(('T', 'Temperature'))

        symbols.append(('volume', 'Zone volume'))
        symbols.append(('fluid_volume', 'Zone fluid volume (different if porous)'))

        for (nme, val) in self.notebook.getNotebookList():
            symbols.append((nme, 'value (notebook) = ' + str(val)))

        return exp, req, self.list_scalars, symbols;


    def getFormulaAlComponents(self, zone="1"):
        """
        User formula for thermal conductivity
        """
        exp = self.getFormula('thermal_conductivity', zone)
        req = [('thermal_conductivity', 'Thermal conductivity')]

        symbols = []
        for s in self.list_scalars:
           symbols.append(s)

        lambda0_value = self.getInitialValueCond()
        ref_pressure = self.getPressure()
        symbols.append(('lambda0', 'Thermal conductivity (reference value) = ' + str(lambda0_value)))
        symbols.append(('p0', 'Reference pressure = ' + str(ref_pressure)))

        symbols.append(('volume', 'Zone volume'))
        symbols.append(('fluid_volume', 'Zone fluid volume (different if porous)'))

        for (nme, val) in self.notebook.getNotebookList():
            symbols.append((nme, 'value (notebook) = ' + str(val)))

        return exp, req, self.list_scalars, symbols;


    def getFormulaDiffComponents(self, scalar, zone="1"):
        """
        User formula for the diffusion coefficient
        """
        name = self.m_sca.getScalarDiffusivityName(scalar)
        exp  = self.m_sca.getDiffFormula(scalar, zone)
        req = [(str(name), str(scalar)+' diffusion coefficient')]
        sym = [('x', 'cell center coordinate'),
               ('y', 'cell center coordinate'),
               ('z', 'cell center coordinate'),
               ('volume', 'Zone volume'),
               ('fluid_volume', 'Zone fluid volume (different if porous)')]
        sym.append((str(scalar),str(scalar)))
        diff0_value = self.m_sca.getScalarDiffusivityInitialValue(scalar)
        sym.append((str(name)+'_ref', str(scalar)+' diffusion coefficient (reference value, m^2/s) = '+str(diff0_value)))

        sym.append(('volume', 'Zone volume'))
        sym.append(('fluid_volume', 'Zone fluid volume (different if porous)'))

        for (nme, val) in self.notebook.getNotebookList():
            sym.append((nme, 'value (notebook) = ' + str(val)))

        return exp, req, self.list_scalars, sym


##    def RemoveThermoConductNode(self):
##        """Remove balise property for thermal_conductivity"""
##        self.node_fluid.xmlRemoveChild('property', name='thermal_conductivity')


    def tr(self, text):
        """
        translation
        """
        return text


#-------------------------------------------------------------------------------
# FluidCharacteristicsModel test case
#-------------------------------------------------------------------------------


class FluidCharacteristicsModelTestCase(ModelTest):
    """
    """
    def checkFluidCharacteristicsInstantiation(self):
        """Check whether the FluidCaracteristicsModel class could be instantiated"""
        model = None
        model = FluidCharacteristicsModel(self.case)
        assert model != None, 'Could not instantiate FluidCaracteristicsModel'

    def checkGetThermoPhysicalModel(self):
        """Check whether thermal physical models could be get"""
        mdl = FluidCharacteristicsModel(self.case)
        from code_saturne.model.ThermalScalarModel import ThermalScalarModel
        ThermalScalarModel(self.case).setThermalModel('temperature_celsius')
        del ThermalScalarModel
        assert mdl.getThermoPhysicalModel() == ('off', 'off', 'temperature_celsius', 'off', 'off', 'off', 'off'),\
        'Could not get thermophysical models in FluidCaracteristicsModel'

    def checkSetandGetInitialValue(self):
        """Check whether the initial value for the properties could be set and get"""
        mdl = FluidCharacteristicsModel(self.case)
        mdl.setInitialValue('density', 123.0)
        mdl.setInitialValue('molecular_viscosity', 1.5e-5)
        mdl.setInitialValue('specific_heat', 1212)
        mdl.setInitialValue('thermal_conductivity', 0.04)
        doc = '''<fluid_properties>
                    <property choice="constant" label="Density" name="density">
                        <initial_value>123</initial_value>
                    </property>
                    <property choice="constant" label="LamVisc" name="molecular_viscosity">
                        <initial_value>1.5e-05</initial_value>
                    </property>
                    <property choice="constant" label="Sp. heat" name="specific_heat">
                        <initial_value>1212</initial_value>
                    </property>
                    <property choice="constant" label="Th. cond" name="thermal_conductivity">
                        <initial_value>0.04</initial_value>
                    </property>
                 </fluid_properties>'''
        assert mdl.node_fluid == self.xmlNodeFromString(doc),\
        'Could not set initial value for properties in FluidCharacteristicsModel model'
        assert mdl.getInitialValue('specific_heat') == 1212.,\
         'Could not get initial value for properties in FluidCharacteristicsModel model'

    def checkSetandGetInitialValueDensity(self):
        """Check whether the initial value for density could be set and get"""
        mdl = FluidCharacteristicsModel(self.case)
        mdl.setInitialValueDensity(89.98)
        doc = '''<property choice="constant" label="Density" name="density">
                    <initial_value>89.98</initial_value>
                 </property>'''
        assert mdl.node_density == self.xmlNodeFromString(doc),\
        'Could not set initial value of density'
        assert mdl.getInitialValueDensity() == 89.98,\
        'Could not get initial value of density'

    def checkSetandGetInitialValueViscosity(self):
        """Check whether the initial value for molecular_viscosity could be set and get"""
        mdl = FluidCharacteristicsModel(self.case)
        mdl.setInitialValueViscosity(1.2e-4)
        doc = '''<property choice="constant" label="LamVisc" name="molecular_viscosity">
                    <initial_value>1.2e-4</initial_value>
                 </property>'''
        assert mdl.node_viscosity == self.xmlNodeFromString(doc),\
        'Could not set initial value of molecular_viscosity'
        assert mdl.getInitialValueViscosity() == 1.2e-4,\
        'Could not get initial value of molecular_viscosity'

    def checkSetandGetInitialValueHeat(self):
        """Check whether the initial value for specific_heat could be set and get"""
        mdl = FluidCharacteristicsModel(self.case)
        mdl.setInitialValueHeat(456.9)
        doc = '''<property choice="constant" label="Sp. heat" name="specific_heat">
                    <initial_value>456.9</initial_value>
                 </property>'''
        assert mdl.node_heat == self.xmlNodeFromString(doc),\
        'Could not set initial value of specific_heat'
        assert mdl.getInitialValueHeat() == 456.9,\
        'Could not get initial value of specific_heat'

    def checkSetandGetInitialValueCond(self):
        """Check whether the initial value for thermal_conductivity could be set and get"""
        mdl = FluidCharacteristicsModel(self.case)
        mdl.setInitialValueCond(456.9)
        doc = '''<property choice="constant" label="Th. cond" name="thermal_conductivity">
                    <initial_value>456.9</initial_value>
                 </property>'''
        assert mdl.node_cond == self.xmlNodeFromString(doc),\
        'Could not set initial value of thermal_conductivity'
        assert mdl.getInitialValueCond() == 456.9,\
        'Could not get initial value of thermal_conductivity'

    def checkSetandGetPropertyMode(self):
        """Check whether choice constant or variable could be set and get"""
        mdl = FluidCharacteristicsModel(self.case)
        mdl.setPropertyMode('density', 'user_law')
        doc = '''<property choice="user_law" label="Density" name="density">
                    <initial_value>1.17862</initial_value>
                 </property>'''
        assert mdl.node_density == self.xmlNodeFromString(doc),\
        'Could not set choice constant or variable for property'
        assert mdl.getPropertyMode('density') == 'user_law',\
        'Could not get choice constant or variable for property'

        mdl.setPropertyMode('density', 'constant')
        doc2 = '''<property choice="constant" label="Density" name="density">
                    <initial_value>1.17862</initial_value>
                 </property>'''
        assert mdl.node_density == self.xmlNodeFromString(doc2),\
        'Could not set listing and recording status for property and remove this nodes'

##    def checkRemoveThermoConductNode(self):
##        """Check whether node thermal conductivity could be removed"""
##        mdl = FluidCharacteristicsModel(self.case)
##        mdl.setPropertyMode('density', 'variable')
##        mdl.setPropertyMode('molecular_viscosity', 'variable')
##        mdl.setPropertyMode('specific_heat', 'variable')
##        mdl.RemoveThermoConductNode()
##        doc = '''<fluid_properties>
##                    <property choice="variable" label="Density" name="density">
##                        <initial_value>1.17862</initial_value>
##                    </property>
##                    <property choice="variable" label="LamVisc" name="molecular_viscosity">
##                        <initial_value>1.83e-05</initial_value>
##                    </property>
##                    <property choice="variable" label="Sp. heat" name="specific_heat">
##                        <initial_value>1017.24</initial_value>
##                    </property>
##                 </fluid_properties>'''
##        assert mdl.node_fluid == self.xmlNodeFromString(doc),\
##        'Could not remove thermal_conductivity node'


def suite():
    testSuite = unittest.makeSuite(FluidCharacteristicsModelTestCase, "check")
    return testSuite


def runTest():
    print(__file__)
    runner = unittest.TextTestRunner()
    runner.run(suite())


#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------



#!/usr/bin/env python3

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

import os

#===============================================================================
# Local functions
#===============================================================================

#===============================================================================
# Defining parameters for a calculation domain
#===============================================================================

def domain_prepare_data_add(domain):
    """
    Additional steps to prepare data
    (called in data preparation stage, between copy of files
    in DATA and copy of link of restart files as defined by domain).
    """

    return

#-------------------------------------------------------------------------------

def domain_copy_results_add(domain):
    """
    Additional steps to copy results or cleanup execution directory
    (called at beginning of results copy stage).
    """

    return

#-------------------------------------------------------------------------------

def define_domain_parameters(domain):
    """
    Define domain execution parameters.
    (called in data preparation stage, just after copy of base setup files
    in execution directory, and before compiling user-defined functions).
    """

    # Reusing output from previous runs
    #----------------------------------

    # By default, upon restart, if the previous mesh is available in the
    # checkpoint, preprocessing is skipped. To allow additional preprocessing,
    # set:
    #   domain.preprocess_on_restart = True

    # To use the output of a previous Preprocessor (mesh import) run, set:
    #   domain.mesh_input = 'RESU/<run_id>/mesh_input'

    # To use the mesh output of a previous solver run, set:
    #   domain.mesh_input = 'RESU/<run_id>/mesh_output'

    # To reuse a previous partitioning, set:
    #   domain.partition_input = 'RESU/<run_id>/partition'

    # To continue a calculation from a previous checkpoint, set:
    #  domain.restart_input = 'RESU/<run_id>/checkpoint'

    # To continue a calculation from a previous checkpoint with
    # a different mesh, it may be necessary to provide a mesh_input
    # matching that checkoint if is not contained, so set:
    #  domain.restart_mesh_input = 'RESU/<run_id>/restart_mesh_input'

    domain.preprocess_on_restart = False
    domain.mesh_input = None
    domain.partition_input = None
    domain.restart_input = None
    domain.restart_mesh_input = None

    # Defining meshes to import (only if domain.mesh_input = None)
    #-------------------------------------------------------------

    # A case-specific mesh directory can be defined through domain.mesh_dir.

    # domain.mesh_dir = None

    # Mesh names can be given as absolute path names; otherwise, file are
    # searched in the following directories, in the order:
    #   - the   case mesh database directory, domain.mesh_dir (if defined)
    #   - the  study mesh database directory, ../MESH directory (if present)
    #   - the   user mesh database directory, see ~/.code_saturne.cfg
    #   - the global mesh database directory, see $prefix/etc/code_saturne.cfg

    # Meshes should be defined as a list of strings defining mesh
    # file names. If preprocessor options must be used for some meshes,
    # a tuple containing a file name followed by additional options
    # may be used instead of a simple file name, for example:
    #   domain.meshes = ['part1.des',
    #                    ('part2_med', '--format', 'med',
    #                     '--num', '2', '--reorient'),
    #                    '~/meshdatabase/part3.unv']

    # domain.meshes = None

    # Logging arguments
    #------------------

    # Command-line arguments useful for logging, or determining the calculation
    # type may be defined here, for example:
    #   domain.logging_args = '--logp'

    # domain.logging_args = None

    # Solver options
    #---------------

    # Running the solver may be deactivated by setting
    #   domain.exec_solver = False

    # The type of run may be determined using domain.solver_args.
    # For example:
    #   domain.solver_args = '--quality'
    # allows activation of elementary mesh quality criteria output, while:
    #   domain.solver_args = '--benchmark'
    # allows running  basic linear algebra operation benchmarks.
    # To run the solver's preprocessing stage only (mesh joining, smoothing,
    # and other modifications), use:
    #   domain.solver_args = '--preprocess'

    # domain.exec_solver = True
    # domain.solver_args = None

    # Compile and build options
    #--------------------------

    # Additionnal compiler flags may be passed to the C, C++, or Fortran
    # compilers, and libraries may be added, in case linking of user
    # functions against external libraries is needed.

    # Note that compiler flags will be added before the default flags;
    # this helps ensure added search paths have priority, but also implies
    # that user optimization options may be superceded by the default ones.

    domain.compile_cflags = None
    domain.compile_cxxflags = None
    domain.compile_fcflags = None
    domain.compile_libs = None

    # Debugging options
    #------------------

    # To run the solver through a debugger, domain.debug should contain
    # the matching command-line arguments, such as:
    #   domain.debug = '--debugger=gdb'
    # or (for Valgrind):
    #   domain.debug = 'valgrind --tool=memcheck'
    # or (for Valgrind and ddd):
    #   domain.debug = '--debugger=ddd valgrind --tool=memcheck --vgdb-error=1'

    # domain.debug = None

    # import pprint
    # pprint.pprint(domain.__dict__)

    return

#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module defines the Cathare coupling model data management.

This module contains the following classes and function:
- CathareCouplingModel
"""

#-------------------------------------------------------------------------------
# Library modules
#-------------------------------------------------------------------------------

import sys, unittest

#-------------------------------------------------------------------------------
# Application modules import
#-------------------------------------------------------------------------------

from code_saturne.model.Common import *
from code_saturne.model.XMLvariables import Variables, Model
from code_saturne.model.XMLmodel import ModelTest

#-------------------------------------------------------------------------------
# Cathare coupling model class
#-------------------------------------------------------------------------------

class CathareCouplingModel(Variables, Model):
    """
    Manage the input/output markups in the xml doc
    """
    def __init__(self, case):
        """
        Constructor.
        """
        self.case = case

        self.__node_models = self.case.xmlGetNode('thermophysical_models')
        self.__node_nepcat = self.__node_models.xmlInitNode('cathare_coupling')

    def defaultValues(self):
        """
        Return in a dictionnary which contains default values.
        """
        default = {}
        default['cathare_elt']        = "TUBE1"
        default['cathare_first_cell'] = 1
        default['cathare_last_cell']  = 2
        default['neptune_bc']         = "off"
        default['neptune_1d_zone']    = "all[]"
        return default

    # ----------------------------------
    def __getNumberOfCathareCoupling(self):
        return len(self.__node_nepcat.xmlGetNodeList('coupling'))

    # ----------------------------------
    def getNumberOfFluids(self):

        nodef = self.__node_models.xmlGetNode('fields')

        np = len(nodef.xmlGetNodeList('field'))

        return np

    # ----------------------------------
    def getCathareActivationStatus(self):

        value = self.__node_nepcat.xmlGetInt('api_type')
        if value == None:
            value = 0

        return value

    # ----------------------------------
    def setCathareFile(self, value):

        self.__node_nepcat.xmlSetData('jdd_name', value)

    def getCathareFile(self):

        value = self.__node_nepcat.xmlGetString('jdd_name')
        if value == None or value == '':
            value = 'cathare_jdd.dat'

        return value

    # ----------------------------------
    def setCplName(self, value):

        self.__node_nepcat.xmlSetData('coupling_name', value)

    def getCplName(self):

        value = self.__node_nepcat.xmlGetString('coupling_name')
        if value == None or value == '':
            value = 'NEPCAT_CPL'

        return value

    # ----------------------------------
    def setCathareTime(self, value):

        self.__node_nepcat.xmlSetData('cathare_init_time', value)

    def getCathareTime(self):

        value = self.__node_nepcat.xmlGetDouble('cathare_init_time')
        if value == None:
            value = 0.0

        return value

    # ----------------------------------
    def setCathareInstanceName(self, name):

        self.__node_nepcat.xmlSetData('cathare_instance_name', name)


    def getCathareInstanceName(self):

        name = self.__node_nepcat.xmlGetString('cathare_instance_name')
        if name == None:
            name = 'CATHARE'

        return name

    # ----------------------------------
    def setNeptuneInstanceName(self, name):

        self.__node_nepcat.xmlSetData('neptune_instance_name', name)


    def getNeptuneInstanceName(self):

        name = self.__node_nepcat.xmlGetString('neptune_instance_name')
        if name == None:
            name = 'NEPTUNE'

        return name


    # ----------------------------------
    def setCplTime(self, value):

        self.__node_nepcat.xmlSetData('coupling_run_time', value)

    def getCplTime(self):

        value = self.__node_nepcat.xmlGetDouble('coupling_run_time')
        if value == None:
            value = 0.0

        return value

    @Variables.noUndo
    def getCathareCouplingList(self):
        """
        @return: list of Cathare coupling description.
        @rtype: C{List}
        """
        node_list = self.__node_nepcat.xmlGetNodeList('coupling')
        lst = []
        for index in range(len(node_list)):
            num = index + 1

            cathare_elt        = self.getCathareEltName(num)
            cathare_first_cell = self.getCathareFCell(num)
            cathare_last_cell  = self.getCathareLCell(num)
            neptune_bc         = self.getNeptuneBc(num)
            neptune_1d_zone    = self.getNeptune1dZone(num)

            lst.append([cathare_elt, cathare_first_cell, cathare_last_cell,
                        neptune_bc, neptune_1d_zone])

        return lst


    @Variables.undoGlobal
    def addCathareCoupling(self, cathare_elt,
                           cathare_first_cell, cathare_last_cell,
                           neptune_bc, neptune_1d_zone):

        """
        Add a new definition of a Cathare coupling.

        @type cathare_elt: C{String}
        @param cathare_elt: Cathare coupled element
        @type cathare_first_cell: C{Int}
        @param cathare_first_cell: first coupled cell for cathare
        @type cathare_last_cell: C{Int}
        @param cathare_last_cell: last coupled cell for cathare
        @type neptune_bc : C{String}
        @param neptune_bc: neptune coupled boundary name
        @type neptune_1d_zone: C{String}
        @param neptune_1d_zone: selection criteria
        @return: new number of Cathare coupling
        @rtype: C{Int}
        """
        num = len(self.__node_nepcat.xmlGetNodeList('coupling'))
        node_new = self.__node_nepcat.xmlAddChild('coupling')

        num = num + 1
        self.setCathareEltName(num, cathare_elt)
        self.setCathareFCell(num, cathare_first_cell)
        self.setCathareLCell(num, cathare_last_cell)
        self.setNeptuneBc(num, neptune_bc)
        self.setNeptune1dZone(num, neptune_1d_zone)

        return num


    @Variables.undoLocal
    def deleteCathareCoupling(self, num):
        """
        Delete a definition of a Cathare coupling.

        @type num: C{Int}
        @param num: Cathare coupling number
        """
        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        node_list = self.__node_nepcat.xmlGetNodeList('coupling')
        node = node_list[num-1]
        node.xmlRemoveNode()

    #------------------------------------------------------------------
    # Helper function
    #------------------------------------------------------------------
    def __getStringData(self, index, name, setFunction):
        """
        Get string value from xml file.
        """
        self.isLowerOrEqual(index+1, self.__getNumberOfCathareCoupling())
        node = self.__node_nepcat.xmlGetNodeList('coupling')[index]
        value = node.xmlGetString(name)
        return self.__getDefaultDataIfNone(index, value, name, setFunction)


    def __getIntData(self, index, name, setFunction):
        """
        Get int value from xml file.
        """
        self.isLowerOrEqual(index+1, self.__getNumberOfCathareCoupling())
        node = self.__node_nepcat.xmlGetNodeList('coupling')[index]
        value = node.xmlGetInt(name)
        return self.__getDefaultDataIfNone(index, value, name, setFunction)


    def __getDefaultDataIfNone(self, index, value, name, setFunction):
        """
        Get default value if value is none.
        """
        if value == None or value == "":
            value = self.defaultValues()[name]
            setFunction(index+1, value)
        return value

    #------------------------------------------------------------------
    # API type: Stand alone NCFD, coupled NCFD or coupled CATHARE
    #------------------------------------------------------------------
    @Variables.undoLocal
    def setApiType(self, value):

        self.__node_nepcat.xmlSetData('api_type', value)

    @Variables.noUndo
    def getApiType(self):

        return self.__node_nepcat.xmlGetInt('api_type')

    #------------------------------------------------------------------
    # Number of coupled phases
    #------------------------------------------------------------------
    @Variables.undoLocal
    def setNphases(self, value):

        self.__node_nepcat.xmlSetData('nphases', value)

    @Variables.noUndo
    def getNphases(self):

        return self.__node_nepcat.xmlGetInt('nphases')

    #------------------------------------------------------------------
    # Cathare coupled element name
    #------------------------------------------------------------------
    @Variables.undoLocal
    def setCathareEltName(self, num, value):
        """
        Set value of Cathare element name.

        @type num: C{Int}
        @param num: Cathare coupling number
        @type value: C{String}
        @param value: Cathare element name
        """
        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        self.isStr(value)
        node = self.__node_nepcat.xmlGetNodeList('coupling')[num-1]
        node.xmlSetData('cathare_element', value)


    @Variables.noUndo
    def getCathareEltName(self, num):
        """
        Get value of Cathare element name.

        @type num: C{Int}
        @param num: Cathare coupling number
        @return: Cathare element name
        @rtype: C{String}
        """
        return self.__getStringData(num-1,
                                    'cathare_element',
                                    self.setCathareEltName)

    #------------------------------------------------------------------
    # Coupled cells in the Cathare element
    #------------------------------------------------------------------
    @Variables.undoLocal
    def setCathareFCell(self, num, value):
        """
        Set value of Cathare first coupled cell.

        @type num: C{Int}
        @param num: Cathare coupling number
        @type value: C{Int}
        @param value: Cathare first coupled cell
        """
        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        self.isInt(value)
        self.isGreaterOrEqual(value, 1)
        node = self.__node_nepcat.xmlGetNodeList('coupling')[num-1]
        node.xmlSetData('first_cell', value)


    @Variables.noUndo
    def getCathareFCell(self, num):
        """
        Get value of Cathare first coupled cell.

        @type num: C{Int}
        @param num: Cathare coupling number
        @return: Cathare first coupled cell
        @rtype: C{Int}
        """
        return self.__getIntData(num-1,
                                 'first_cell',
                                 self.setCathareFCell)

    #------------------------------------------------------------------
    # Coupled cells in the Cathare element
    #------------------------------------------------------------------
    @Variables.undoLocal
    def setCathareLCell(self, num, value):
        """
        Set value of Cathare last coupled cell.

        @type num: C{Int}
        @param num: Cathare coupling number
        @type value: C{Int}
        @param value: Cathare last coupled cell
        """
        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        self.isInt(value)
        self.isGreaterOrEqual(value, 1)
        node = self.__node_nepcat.xmlGetNodeList('coupling')[num-1]
        node.xmlSetData('last_cell', value)


    @Variables.noUndo
    def getCathareLCell(self, num):
        """
        Get value of Cathare last coupled cell.

        @type num: C{Int}
        @param num: Cathare coupling number
        @return: Cathare last coupled cell
        @rtype: C{Int}
        """
        return self.__getIntData(num-1,
                                 'last_cell',
                                 self.setCathareLCell)

    #------------------------------------------------------------------
    # Coupled Boundary condition
    #------------------------------------------------------------------
    @Variables.undoLocal
    def setNeptuneBc(self, num, value):
        """
        Set value of Neptune coupled boundary name.

        @type num: C{Int}
        @param num: Cathare coupling number
        @type value: C{String}
        @param value: Neptune coupled boundary name
        """
        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        self.isStr(value)
        node = self.__node_nepcat.xmlGetNodeList('coupling')[num-1]
        node.xmlSetData('neptune_bc', value)


    @Variables.noUndo
    def getNeptuneBc(self, num):
        """
        Get value of Neptune coupled boundary name.

        @type num: C{Int}
        @param num: Cathare coupling number
        @return: Neptune coupled boundary name
        @rtype: C{String}
        """
        return self.__getStringData(num-1,
                                    'neptune_bc',
                                    self.setNeptuneBc)

    @Variables.undoLocal
    def setNeptune2dZone(self, num, value):

        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        self.isStr(value)
        node = self.__node_nepcat.xmlGetNodeList('coupling')[num-1]
        node.xmlSetData('selection_criteria_2d', value)


    @Variables.undoLocal
    def setNeptuneBcType(self, num, value):

        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        self.isStr(value)
        node = self.__node_nepcat.xmlGetNodeList('coupling')[num-1]

        bc_type = -1
        if value == 'inlet':
            bc_type = 0
        elif value == 'outlet':
            bc_type = 1
        elif value == 'wall':
            bc_type = 2

        node.xmlSetData('bc_cpl_type', bc_type)


    #------------------------------------------------------------------
    # Selection criteria
    #------------------------------------------------------------------
    @Variables.undoLocal
    def setNeptune1dZone(self, num, value):
        """
        Set value of equivalent coupled 1D zone.

        @type num: C{Int}
        @param num: Cathare coupling number
        @type value: C{String}
        @param value: equivalent coupled 1D zone
        """
        self.isLowerOrEqual(num, self.__getNumberOfCathareCoupling())
        self.isStr(value)
        node = self.__node_nepcat.xmlGetNodeList('coupling')[num-1]
        node.xmlSetData('selection_criteria_3d', value)


    @Variables.noUndo
    def getNeptune1dZone(self, num):
        """
        Get value of equivalent coupled 1D zone.

        @type num: C{Int}
        @param num: Cathare coupling number
        @return: equivalent coupled 1D zone
        @rtype: C{String}
        """
        return self.__getStringData(num-1,
                                    'selection_criteria_3d',
                                    self.setNeptune1dZone)

#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module defines the Page for the physical properties of the fluid.
These properties can be reference value or initial value

This module contains the following classes and function:
- FluidCharacteristicsView
"""

#-------------------------------------------------------------------------------
# Library modules import
#-------------------------------------------------------------------------------

import logging

#-------------------------------------------------------------------------------
# Third-party modules
#-------------------------------------------------------------------------------

from code_saturne.Base.QtCore    import *
from code_saturne.Base.QtGui     import *
from code_saturne.Base.QtWidgets import *

#-------------------------------------------------------------------------------
# Application modules import
#-------------------------------------------------------------------------------

from code_saturne.model.Common import GuiParam
from code_saturne.Base.QtPage import DoubleValidator, ComboModel, from_qvariant
from code_saturne.Pages.FluidCharacteristicsForm import Ui_FluidCharacteristicsForm
from code_saturne.model.AtmosphericFlowsModel import AtmosphericFlowsModel
from code_saturne.model.FluidCharacteristicsModel import FluidCharacteristicsModel
from code_saturne.model.DefineUserScalarsModel import DefineUserScalarsModel
from code_saturne.model.ThermalScalarModel import ThermalScalarModel
from code_saturne.model.GroundwaterModel import GroundwaterModel
from code_saturne.Pages.QMegEditorView import QMegEditorView
from code_saturne.model.NotebookModel import NotebookModel
from code_saturne.model.InternalCouplingModel import InternalCouplingModel
from code_saturne.model.LocalizationModel import LocalizationModel

#-------------------------------------------------------------------------------
# log config
#-------------------------------------------------------------------------------

logging.basicConfig()
log = logging.getLogger("FluidCharacteristicsView")
log.setLevel(GuiParam.DEBUG)

#-------------------------------------------------------------------------------
# Main class
#-------------------------------------------------------------------------------

class FluidCharacteristicsView(QWidget, Ui_FluidCharacteristicsForm):

    """
    Class to open Molecular Properties Page.
    """
    density = """# Density of air
density = 1.293*(273.15 / temperature);


# density for mixtures of gases
#
# Y1 -> mass fraction of component 1
# Y2 -> mass fraction of component 2

rho1 = 1.25051;
rho2 = 1.7832;
A = (Y1 / rho1) + (Y2 /rho2);
density = 1.0 / A;
"""

    density_h = """# Density
density = enthalpy / 1040. * 1.29;

# density for mixtures of gases
#
# Y1 -> mass fraction of component 1
# Y2 -> mass fraction of component 2

rho1 = 1.25051;
rho2 = 1.7832;
A = (Y1 / rho1) + (Y2 /rho2);
density = 1.0 / A;
"""

    density_wo = """density = 1.25051;

"""
    molecular_viscosity="""# Sutherland's Formula
# Gas             Cst    T0      mu0
# air             120    291.15  18.27e-6
# nitrogen        111    300.55  17.81e-6
# oxygen          127    292.25  20.18e-6
# carbon dioxide  240    293.15  14.8e-6
# carbon monoxide 118    288.15  17.2e-6
# hydrogen        72     293.85  8.76e-6
# ammonia         370    293.15  9.82e-6
# sulfur dioxide  416    293.65  12.54e-6
# helium          79.4   273     19e-6

CST = 120;
T0 = 291.15;
mu_ref = 18.27e-6;

if ( temperature > 0 && temperature < 555) {
molecular_viscosity = mu_ref * ((T0+CST) / (temperature+CST)) * (temperature/T0)^(3./2.);
} else {
molecular_viscosity = -999.0;
}
"""

    molecular_viscosity_h="""CST = 120;
T0 = 291.15;
mu_ref = 18.27e-6;
temperature = enthalpy / 1040.;

if ( enthalpy > 0) {
molecular_viscosity = mu_ref * (T0+CST / temperature+CST) * (temperature/T0)^(3./2.);
} else {
molecular_viscosity = -999.0;
}
"""

    molecular_viscosity_wo="""CST = 120;
T0 = 291.15;
mu_ref = 18.27e-6;
molecular_viscosity = mu_ref * (T0+CST);
"""

    specific_heat="""# specific heat for mixtures of gases
#
# Y1 -> mass fraction of component 1
# Y2 -> mass fraction of component 2

Cp1 = 520.3;
Cp2 = 1040.0;
specific_heat = Y1 * Cp1 + Y2 *Cp2;
"""

    volume_viscosity="""# volume_viscosity
"""

    thermal_conductivity="""# oxygen
thermal_conductivity = 6.2e-5 * temperature + 8.1e-3;

# nitrogen
thermal_conductivity = 6.784141e-5 * temperature + 5.564317e-3;

# hydrogen
thermal_conductivity = 4.431e-4 * temperature + 5.334e-2;
"""

    thermal_conductivity_h="""temperature = enthalpy / 1040.;
thermal_conductivity = 6.2e-5 * temperature + 8.1e-3;
"""

    def __init__(self, parent=None): #, case):
        """
        Constructor
        """
        QWidget.__init__(self, parent)

        Ui_FluidCharacteristicsForm.__init__(self)
        self.setupUi(self)

        self.case     = None
        self.mdl      = None
        self.notebook = None
        self.zone     = None
        self.zone_name = None
        self.zone_id   = None
        self.is_main_zone = False
        self.is_solid  = False


    def setup(self, case, zone_name):

        self.case = case

        for zone in LocalizationModel('VolumicZone', self.case).getZones():
            if zone.getLabel() == zone_name:
                self.zone = zone
                self.zone_name = zone.getLabel()
                self.zone_id   = str(zone.getCodeNumber())
                if self.zone_name == "all_cells":
                    self.is_main_zone = True
                elif not zone.isNatureActivated("physical_properties"):
                    return
                if zone.isNatureActivated("solid"):
                    self.is_solid = True

        self.case.undoStopGlobal()

        self.mdl = FluidCharacteristicsModel(self.case)
        self.notebook = NotebookModel(self.case)

        self.internal_cpl = InternalCouplingModel(self.case)
        self.m_th = ThermalScalarModel(self.case)
        s = self.m_th.getThermalScalarName()
        tsm = self.mdl.tsm

        # Particular Widget init. taking into account chosen fluid model
        mdl_atmo, mdl_joule, mdl_thermal, mdl_gas, mdl_coal, mdl_comp, mdl_hgn=\
            self.mdl.getThermoPhysicalModel()

        # Combo models

        self.modelRho       = ComboModel(self.comboBoxRho,      2, 1)
        self.modelMu        = ComboModel(self.comboBoxMu,       2, 1)
        self.modelCp        = ComboModel(self.comboBoxCp,       2, 1)
        self.modelAl        = ComboModel(self.comboBoxAl,       2, 1)
        self.modelDiff      = ComboModel(self.comboBoxDiff,     2, 1)
        self.modelNameDiff  = ComboModel(self.comboBoxNameDiff, 1, 1)
        self.modelViscv0    = ComboModel(self.comboBoxViscv0,   3, 1)
        self.modelDiftl0    = ComboModel(self.comboBoxDiftl0,   3, 1)
        self.modelMaterial  = ComboModel(self.comboBoxMaterial, 1, 1)
        self.modelMethod    = ComboModel(self.comboBoxMethod,   1, 1)
        self.modelReference = ComboModel(self.comboBoxReference,  1, 1)

        if mdl_joule == 'off':
            self.modelRho.addItem(self.tr('constant'), 'constant')
            self.modelRho.addItem(self.tr('material law'), 'thermal_law')
        if mdl_atmo != 'off':
            self.modelRho.addItem(self.tr('predefined law'), 'predefined_law')
        elif mdl_hgn != "off":
            self.modelRho.addItem(self.tr('linear mixture law'), 'predefined_law')
        elif mdl_joule != 'off':
            self.modelRho.addItem(self.tr('predefined law'), 'predefined_law')
        elif mdl_comp != 'off':
            self.modelRho.addItem(self.tr('predefined law'), 'predefined_law')
        elif mdl_gas != 'off' or mdl_coal != 'off':
            self.modelRho.addItem(self.tr('predefined law'), 'predefined_law')
        self.modelRho.addItem(self.tr('user law'), 'user_law')

        if mdl_joule == 'off':
            self.modelMu.addItem(self.tr('constant'), 'constant')
            self.modelMu.addItem(self.tr('material law'), 'thermal_law')
        if mdl_joule != 'off':
            self.modelMu.addItem(self.tr('predefined law'), 'predefined_law')
        if mdl_hgn != "off":
            self.modelMu.addItem(self.tr('linear mixture law'), 'predefined_law')
        self.modelMu.addItem(self.tr('user law'), 'user_law')

        if mdl_joule == 'off':
            self.modelCp.addItem(self.tr('constant'), 'constant')
            self.modelCp.addItem(self.tr('material law'), 'thermal_law')
        if mdl_joule != 'off':
            self.modelCp.addItem(self.tr('predefined law'), 'predefined_law')
        self.modelCp.addItem(self.tr('user law'), 'user_law')

        if mdl_joule == 'off':
            self.modelAl.addItem(self.tr('constant'), 'constant')
            self.modelAl.addItem(self.tr('material law'), 'thermal_law')
        if mdl_joule != 'off':
            self.modelAl.addItem(self.tr('predefined law'), 'predefined_law')
        self.modelAl.addItem(self.tr('user law'), 'user_law')

        self.modelDiff.addItem(self.tr('constant'), 'constant')
        self.modelDiff.addItem(self.tr('user law'), 'user_law')

        self.modelViscv0.addItem(self.tr('constant'), 'constant')
        self.modelViscv0.addItem(self.tr('user law'), 'user_law')
        self.modelViscv0.addItem(self.tr('material law'), 'thermal_law')

        self.modelDiftl0.addItem(self.tr('constant'), 'constant')
        self.modelDiftl0.addItem(self.tr('user law'), 'user_law')
        self.modelDiftl0.addItem(self.tr('material law'), 'thermal_law')
        if mdl_gas != 'off' or mdl_coal != 'off':
            self.modelDiftl0.addItem(self.tr('predefined law'), 'predefined_law')

        self.scalar = ""
        scalar_list = self.mdl.m_sca.getUserScalarNameList()
        for s in self.mdl.m_sca.getScalarsVarianceList():
            if s in scalar_list: scalar_list.remove(s)

        if scalar_list != []:
            self.scalar = scalar_list[0]
            for scalar in scalar_list:
                self.modelNameDiff.addItem(scalar)

        # Validators

        validatorP0 = DoubleValidator(self.lineEditP0, min=0.0)
        self.lineEditP0.setValidator(validatorP0)

        validatorT0 = DoubleValidator(self.lineEditT0,  min=0.0)
        validatorT0.setExclusiveMin(True)
        self.lineEditT0.setValidator(validatorT0)

        validatorOxydant = DoubleValidator(self.lineEditOxydant,  min=0.0)
        validatorOxydant.setExclusiveMin(True)
        self.lineEditOxydant.setValidator(validatorOxydant)

        validatorFuel = DoubleValidator(self.lineEditFuel,  min=0.0)
        validatorFuel.setExclusiveMin(True)
        self.lineEditFuel.setValidator(validatorFuel)

        validatorMM = DoubleValidator(self.lineEditMassMolar, min=0.0)
        validatorMM.setExclusiveMin(True)
        self.lineEditMassMolar.setValidator(validatorMM)

        validatorRho    = DoubleValidator(self.lineEditRho,    min = 0.0)
        validatorRho1   = DoubleValidator(self.lineEditRho1,   min = 0.0)
        validatorRho2   = DoubleValidator(self.lineEditRho2,   min = 0.0)
        validatorMu     = DoubleValidator(self.lineEditMu,     min = 0.0)
        validatorMu1    = DoubleValidator(self.lineEditMu1,    min = 0.0)
        validatorMu2    = DoubleValidator(self.lineEditMu2,    min = 0.0)
        validatorCp     = DoubleValidator(self.lineEditCp,     min = 0.0)
        validatorAl     = DoubleValidator(self.lineEditAl,     min = 0.0)
        validatorDiff   = DoubleValidator(self.lineEditDiff,   min = 0.0)
        validatorViscv0 = DoubleValidator(self.lineEditViscv0, min = 0.0)
        validatorDiftl0 = DoubleValidator(self.lineEditDiftl0, min = 0.0)

        validatorRho.setExclusiveMin(True)
        validatorRho1.setExclusiveMin(True)
        validatorRho2.setExclusiveMin(True)
        validatorMu.setExclusiveMin(True)
        validatorMu1.setExclusiveMin(True)
        validatorMu2.setExclusiveMin(True)
        validatorCp.setExclusiveMin(True)
        validatorAl.setExclusiveMin(True)
        validatorDiff.setExclusiveMin(True)
        validatorDiftl0.setExclusiveMin(True)

        self.lineEditRho.setValidator(validatorRho)
        self.lineEditRho1.setValidator(validatorRho1)
        self.lineEditRho2.setValidator(validatorRho2)
        self.lineEditMu.setValidator(validatorMu)
        self.lineEditMu1.setValidator(validatorMu1)
        self.lineEditMu2.setValidator(validatorMu2)
        self.lineEditCp.setValidator(validatorCp)
        self.lineEditAl.setValidator(validatorAl)
        self.lineEditDiff.setValidator(validatorDiff)
        self.lineEditViscv0.setValidator(validatorViscv0)
        self.lineEditDiftl0.setValidator(validatorDiftl0)

        # Connections

        self.lineEditP0.textChanged[str].connect(self.slotPressure)
        self.lineEditT0.textChanged[str].connect(self.slotTemperature)
        self.lineEditOxydant.textChanged[str].connect(self.slotTempOxydant)
        self.lineEditFuel.textChanged[str].connect(self.slotTempFuel)
        self.lineEditMassMolar.textChanged[str].connect(self.slotMassemol)

        self.comboBoxRho.currentIndexChanged[str].connect(self.slotStateRho)
        self.comboBoxMu.currentIndexChanged[str].connect(self.slotStateMu)
        self.comboBoxCp.currentIndexChanged[str].connect(self.slotStateCp)
        self.comboBoxAl.currentIndexChanged[str].connect(self.slotStateAl)
        self.comboBoxDiff.currentIndexChanged[str].connect(self.slotStateDiff)
        self.comboBoxNameDiff.activated[str].connect(self.slotNameDiff)
        self.comboBoxViscv0.currentIndexChanged[str].connect(self.slotStateViscv0)
        self.comboBoxMaterial.activated[str].connect(self.slotMaterial)
        self.comboBoxMethod.activated[str].connect(self.slotMethod)
        self.comboBoxReference.activated[str].connect(self.slotReference)
        self.lineEditRho.textChanged[str].connect(self.slotRho)
        self.lineEditRho1.textChanged[str].connect(self.slotRho1)
        self.lineEditRho2.textChanged[str].connect(self.slotRho2)
        self.lineEditMu.textChanged[str].connect(self.slotMu)
        self.lineEditMu1.textChanged[str].connect(self.slotMu1)
        self.lineEditMu2.textChanged[str].connect(self.slotMu2)
        self.lineEditCp.textChanged[str].connect(self.slotCp)
        self.lineEditAl.textChanged[str].connect(self.slotAl)
        self.lineEditDiff.textChanged[str].connect(self.slotDiff)
        self.lineEditDiftl0.textChanged[str].connect(self.slotDiftl0)
        self.lineEditViscv0.textChanged[str].connect(self.slotViscv0)
        self.pushButtonRho.clicked.connect(self.slotFormulaRho)
        self.pushButtonMu.clicked.connect(self.slotFormulaMu)
        self.pushButtonCp.clicked.connect(self.slotFormulaCp)
        self.pushButtonAl.clicked.connect(self.slotFormulaAl)
        self.pushButtonDiff.clicked.connect(self.slotFormulaDiff)
        self.pushButtonViscv0.clicked.connect(self.slotFormulaViscv0)

        self.initializeWidget()

        self.case.undoStartGlobal()


    def initializeWidget(self):
        """
        """
        mdls = self.mdl.getThermoPhysicalModel()
        mdl_atmo, mdl_joule, mdl_thermal, mdl_gas, mdl_coal, mdl_comp, mdl_hgn = mdls

        if self.is_solid:
            mdl_atmo = "off"
            mdl_joule = "off"
            mdl_gas = "off"
            mdl_coal = "off"
            mdl_comp = "off"
            mdl_hgn = "off"

        is_main_zone = self.is_main_zone

        self.groupBoxPressure.setVisible(is_main_zone)
        self.groupBoxTemperature.setVisible(is_main_zone)

        self.groupBoxMassMolar.hide()

        isLargeScaleMeteoChecked = AtmosphericFlowsModel(self.case).getLargeScaleMeteoStatus() == 'on'
        if mdl_atmo != "off" and isLargeScaleMeteoChecked:
            self.groupBoxTemperature.setEnabled(False)
            t = AtmosphericFlowsModel(self.case).getMeteoT0()
            self.lineEditT0.setText(str(t))

        elif mdl_comp != "off" or mdl_coal != "off":
            m = self.mdl.getMassemol()
            self.lineEditMassMolar.setText(str(m))
            self.groupBoxMassMolar.setVisible(is_main_zone)
            if mdl_comp != "off":
                t = self.mdl.getTemperature()
                self.lineEditT0.setText(str(t))

        elif mdl_thermal != "off" or mdl_gas == 'd3p':
            t = self.mdl.getTemperature()
            self.lineEditT0.setText(str(t))
            if mdl_thermal == "temperature_celsius":
                self.labelUnitT0.setText("\xB0 C")

            if self.mdl.getMaterials() != "user_material":
                self.labelInfoT0.hide()
        else:
            self.groupBoxTemperature.hide()

        if mdl_gas == 'd3p':
            self.groupBoxTempd3p.setVisible(is_main_zone)
            t_oxy  = self.mdl.getTempOxydant()
            t_fuel = self.mdl.getTempFuel()
            self.lineEditOxydant.setText(str(t_oxy))
            self.lineEditFuel.setText(str(t_fuel))
        else:
            self.groupBoxTempd3p.hide()

        darc = GroundwaterModel(self.case).getGroundwaterModel()
        if darc != 'off':
            self.groupBoxPressure.hide()
        elif mdl_atmo != 'off' and isLargeScaleMeteoChecked:
            self.groupBoxPressure.setEnabled(False)
            p = AtmosphericFlowsModel(self.case).getMeteoPsea()
            self.lineEditP0.setText(str(p))
            self.groupBoxPressure.setEnabled(False)

        else:
            p = self.mdl.getPressure()
            self.lineEditP0.setText(str(p))

        if self.mdl.tables == 0 or mdl_joule != 'off' or mdl_comp != 'off':
            self.groupBoxTableChoice.hide()
        else:
            self.groupBoxTableChoice.setVisible(is_main_zone)

            fluids = self.mdl.getLibPropertiesDict()
            o_keys = list(fluids.keys()) # for Python2/3 compatibility
                                         # fluids.keys() ordered and iteratable
                                         # already in Python3
            o_keys.sort()
            for f in o_keys:
                if fluids[f] != 0:
                    self.modelMaterial.addItem(self.tr(f), f)
                else:
                    self.modelMaterial.addItem(self.tr(f), True)

            material = self.mdl.getMaterials()
            self.modelMaterial.setItem(str_model=material)
            self.updateMethod()

        # VoF
        if mdl_hgn != 'off':
            self.widgetRefRho.hide()
            self.widgetVofRho.setVisible(is_main_zone)
            self.widgetRefMu.hide()
            self.widgetVofMu.setVisible(is_main_zone)
        elif mdl_atmo != 'off' and isLargeScaleMeteoChecked:
            self.widgetRefRho.hide()
            self.widgetVofRho.hide()
            self.widgetRefMu.setVisible(is_main_zone)
            self.widgetVofMu.hide()
        else:
            if mdl_comp != "off":
                self.widgetRefRho.hide()
            else:
                self.widgetRefRho.setVisible(is_main_zone)
            self.widgetVofRho.hide()
            self.widgetRefMu.setVisible(is_main_zone)
            self.widgetVofMu.hide()

        # compressible
        self.groupBoxViscv0.hide()

        # combustion
        self.groupBoxDiftl0.hide()

        if self.scalar == "":
            self.groupBoxDiff.hide()
        else :
            self.groupBoxDiff.show()
            self.lineEditDiff.setText(str(self.mdl.m_sca.getScalarDiffusivityInitialValue(self.scalar)))

            diff_choice =  self.mdl.m_sca.getScalarDiffusivityChoice(self.scalar)
            self.modelDiff.setItem(str_model=diff_choice)
            self.modelNameDiff.setItem(str_model=str(self.scalar))
            if diff_choice  != 'user_law':
                self.pushButtonDiff.hide()
                self.pushButtonDiff.setEnabled(False)
                self.pushButtonDiff.setStyleSheet("background-color: None")
            else:
                self.pushButtonDiff.show()
                self.pushButtonDiff.setEnabled(True)
                name = self.mdl.m_sca.getScalarDiffusivityName(self.scalar)
                exp = self.mdl.m_sca.getDiffFormula(self.scalar, self.zone_id)
                if exp:
                    self.pushButtonDiff.setStyleSheet("background-color: green")
                    self.pushButtonDiff.setToolTip(exp)
                else:
                    self.pushButtonDiff.setStyleSheet("background-color: red")

            self.lineEditDiff.setVisible(is_main_zone)
            self.labelDiff.setVisible(is_main_zone)
            self.labelUnitDiff.setVisible(is_main_zone)
            self.comboBoxDiff.setEnabled(is_main_zone)

        # Solid zone

        if self.is_solid:
            self.groupBoxTempd3p.hide()
            self.groupBoxMu.hide()
            self.groupBoxDiftl0.hide()
            self.groupBoxDiff.hide()

        # Standard Widget initialization
        for tag, symbol in self.mdl.lst:
            __model  = getattr(self, "model"      + symbol)
            __line   = getattr(self, "lineEdit"   + symbol)
            __button = getattr(self, "pushButton" + symbol)
            __label  = getattr(self, "label"      + symbol)
            __labelu = getattr(self, "labelUnit"  + symbol)

            try:
                __combo  = getattr(self, "comboBox" + symbol)
                __combo.setEnabled(is_main_zone)
            except:
                pass

            if tag != 'dynamic_diffusion':
                __labelv = getattr(self, "labelVar"   + symbol)
                c = self.mdl.getPropertyMode(tag)
                if c not in __model.getItems():
                    c = 'constant'
                    self.mdl.setPropertyMode(tag, c)
                if self.is_solid:
                    c = "user_law"

                __model.setItem(str_model=c)
                if c == 'user_law':
                    __button.show()
                    __button.setEnabled(True)
                    __label.setText(self.tr("Reference value"))
                else:
                    __button.hide()
                    __button.setEnabled(False)
                    __label.setText(self.tr("Reference value"))
                if c == 'thermal_law':
                    __line.hide()
                    __label.hide()
                    __labelu.hide()
                    __labelv.hide()
                else:
                    __line.setVisible(is_main_zone)
                    __label.setVisible(is_main_zone)
                    __labelu.setVisible(is_main_zone)
                    __labelv.setVisible(is_main_zone)
                try:
                    if self.mdl.getMaterials() == "user_material":
                        __model.disableItem(str_model='thermal_law')
                    else:
                        __model.enableItem(str_model='thermal_law')
                except Exception:
                    pass

            else:
                __label.setText(self.tr("Reference value"))

            __line.setText(str(self.mdl.getInitialValue(tag)))


        # If no thermal scalar, hide specific heat and thermal conductivity.
        if mdl_thermal == 'off':
            self.groupBoxCp.hide()
            self.groupBoxAl.hide()

        if mdl_gas != 'off' or mdl_coal != 'off':
            self.groupBoxDiftl0.setVisible(is_main_zone)

        for tag, symbol in self.mdl.lst:
            __model  = getattr(self, "model" + symbol)
            __line   = getattr(self, "lineEdit" + symbol)
            __button = getattr(self, "pushButton" + symbol)
            __label  = getattr(self, "label" + symbol)
            __combo  = getattr(self, "comboBox" + symbol)

            # Gas or coal combustion
            if mdl_gas != 'off' or mdl_coal != 'off':
                if tag == 'density':
                    __model.setItem(str_model='predefined_law')
                    __combo.setEnabled(False)
                    __button.setEnabled(False)
                    __button.hide()
                    self.mdl.setPropertyMode(tag, 'predefined_law')
                    __label.setText(self.tr("Calculation by\n perfect gas law"))
                    __line.setText(str(""))
                    __line.setEnabled(False)
                elif tag == 'dynamic_diffusion':
                    __model.setItem(str_model='predefined_law')
                    __combo.setEnabled(False)
                    __button.setEnabled(False)
                    __button.hide()
                else:
                    choice_p = self.mdl.getPropertyMode(tag)
                    __model.setItem(str_model=choice_p)

            # Joule
            if mdl_joule == 'arc':
                __model.setItem(str_model='predefined_law')
                __combo.setEnabled(False)
                __button.setEnabled(False)
                __button.hide()
                self.mdl.setPropertyMode(tag, 'predefined_law')

            # Atmospheric Flows
            if mdl_atmo != 'off':
                if tag == 'density':
                    __model.disableItem(str_model='constant')
                    __model.disableItem(str_model='predefined_law')
                    __model.setItem(str_model='predefined_law')
                    __combo.setEnabled(False)
                    __button.setEnabled(False)
                    __button.hide()

            # VoF
            if mdl_hgn != 'off':
                if tag == 'molecular_viscosity' or tag == 'density':
                    __model.disableItem(str_model='constant')
                    __model.disableItem(str_model='predefined_law')
                    __model.setItem(str_model='predefined_law')
                    __combo.setEnabled(False)
                    __button.setEnabled(False)
                    __button.hide()

            # Compressible Flows
            if mdl_comp != 'off':
                self.groupBoxViscv0.setVisible(is_main_zone)
                if tag == 'density':
                    __model.setItem(str_model='predefined_law')
                    __combo.setEnabled(False)
                    __button.setEnabled(False)
                    __button.hide()
                    self.mdl.setPropertyMode(tag, 'predefined_law')
                    __line.setEnabled(True)
                if tag == 'specific_heat':
                    __model.setItem(str_model='constant')
                    __combo.setEnabled(False)
                    __button.setEnabled(False)
                    __button.hide()
                    self.mdl.setPropertyMode(tag, 'constant')
                    self.groupBoxCp.setTitle('Isobaric specific heat')

                if tag == 'volume_viscosity':
                    __combo.setEnabled(True)
                    c = self.mdl.getPropertyMode(tag)
                    if c == 'user_law':
                        __button.setEnabled(True)
                    else:
                        __button.setEnabled(False)
            else:
                if tag == 'specific_heat':
                    self.groupBoxCp.setTitle('Specific heat')


        if mdl_hgn != 'off':
            self.lineEditRho1.setText(str(self.mdl.getVofValueDensity(0)))
            self.lineEditRho2.setText(str(self.mdl.getVofValueDensity(1)))

            self.lineEditMu1.setText(str(self.mdl.getVofValueViscosity(0)))
            self.lineEditMu2.setText(str(self.mdl.getVofValueViscosity(1)))


    def updateTypeChoice(self, old_choice):
        """
        add/suppress thermo tables for each properties
        """
        for tag, symbol in self.mdl.lst:
            __model  = getattr(self, "model" + symbol)
            if self.mdl.getMaterials() == "user_material":
                __model.disableItem(str_model='thermal_law')
            else:
                __model.enableItem(str_model='thermal_law')
                if old_choice == "user_material":
                    self.mdl.setPropertyMode(tag, 'thermal_law')
                self.__changeChoice(str("material law"), symbol, tag)
            c = self.mdl.getPropertyMode(tag)
            __model.setItem(str_model=c)


    def updateMethod(self):
        """
        update method list with material choice
        """

        for nb in range(len(self.modelMethod.getItems())):
            self.modelMethod.delItem(0)

        material = self.mdl.getMaterials()
        method = self.mdl.getMethod()
        have_method = False

        methods = self.mdl.getLibPropertyMethods(material)
        for m in methods:
            if method == m[0]:
                have_method = True
            self.modelMethod.addItem(self.tr(m[0]), m[0], not m[1])

        # current method not in list, add it with warning
        if not have_method:
            self.modelMethod.addItem(self.tr(method), method, True)

        # update comboBoxMethod
        self.modelMethod.setItem(str_model=method)
        self.mdl.setMethod(method)

        self.updateReference(material, method)


    def updateReference(self, material, method):
        """
        update method list with material choice
        """

        self.comboBoxReference.hide()
        self.labelReference.hide()

        for nb in range(len(self.modelReference.getItems())):
            self.modelReference.delItem(0)

        reference = self.mdl.getReference()
        references = self.mdl.getAvailReferences(material, method)
        have_reference = False

        if references:
            for r in references:
                if reference == r:
                    have_reference = True
                self.modelReference.addItem(self.tr(r), r)
            self.comboBoxReference.show()
            self.labelReference.show()

            if not have_reference:
                reference = references[0]
            self.modelReference.setItem(str_model=reference)
            self.mdl.setReference(reference)


    @pyqtSlot(str)
    def slotPressure(self,  text):
        """
        Input PRESS.
        """
        if self.lineEditP0.validator().state == QValidator.Acceptable:
            p = from_qvariant(text, float)
            self.mdl.setPressure(p)


    @pyqtSlot(str)
    def slotTemperature(self,  text):
        """
        Input TEMPERATURE.
        """
        if self.lineEditT0.validator().state == QValidator.Acceptable:
            t = from_qvariant(text, float)
            self.mdl.setTemperature(t)


    @pyqtSlot(str)
    def slotTempOxydant(self,  text):
        """
        Input oxydant TEMPERATURE.
        """
        if self.lineEditOxydant.validator().state == QValidator.Acceptable:
            t = from_qvariant(text, float)
            self.mdl.setTempOxydant(t)


    @pyqtSlot(str)
    def slotTempFuel(self,  text):
        """
        Input fuel TEMPERATURE.
        """
        if self.lineEditFuel.validator().state == QValidator.Acceptable:
            t = from_qvariant(text, float)
            self.mdl.setTempFuel(t)


    @pyqtSlot(str)
    def slotMassemol(self,  text):
        """
        Input Mass molar.
        """
        if self.lineEditMassMolar.validator().state == QValidator.Acceptable:
            m = from_qvariant(text, float)
            self.mdl.setMassemol(m)


    @pyqtSlot(str)
    def slotMaterial(self, text):
        """
        Method to call 'setMaterial'
        """
        choice = self.modelMaterial.dicoV2M[str(text)]
        old_choice = self.mdl.getMaterials()
        self.mdl.setMaterials(choice)
        self.updateMethod()
        self.updateTypeChoice(old_choice)


    @pyqtSlot(str)
    def slotMethod(self, text):
        """
        Method to call 'setMethod'
        """
        choice = self.modelMethod.dicoV2M[str(text)]
        self.mdl.setMethod(choice)
        self.updateReference(self.mdl.getMaterials(), choice)


    @pyqtSlot(str)
    def slotReference(self, text):
        """
        Method to call 'setReference'
        """
        choice = self.modelReference.dicoV2M[str(text)]
        self.mdl.setReference(choice)


    @pyqtSlot(str)
    def slotStateRho(self, text):
        """
        Method to call 'getState' with correct arguements for 'rho'
        """
        self.__changeChoice(str(text), 'Rho', 'density')


    @pyqtSlot(str)
    def slotStateMu(self, text):
        """
        Method to call 'getState' with correct arguements for 'Mu'
        """
        self.__changeChoice(str(text), 'Mu', 'molecular_viscosity')


    @pyqtSlot(str)
    def slotStateCp(self, text):
        """
        Method to call 'getState' with correct arguements for 'Cp'
        """
        self.__changeChoice(str(text), 'Cp', 'specific_heat')


    @pyqtSlot(str)
    def slotStateViscv0(self, text):
        """
        Method to call 'getState' with correct arguements for 'Viscv0'
        """
        self.__changeChoice(str(text), 'Viscv0', 'volume_viscosity')


    @pyqtSlot(str)
    def slotStateAl(self, text):
        """
        Method to call 'getState' with correct arguements for 'Al'
        """
        self.__changeChoice(str(text), 'Al', 'thermal_conductivity')


    @pyqtSlot(str)
    def slotStateDiff(self, text):
        """
        Method to set diffusion choice for the coefficient
        """
        choice = self.modelDiff.dicoV2M[str(text)]
        log.debug("slotStateDiff -> %s" % (text))

        if choice != 'user_law':
            self.pushButtonDiff.setEnabled(False)
            self.pushButtonDiff.setStyleSheet("background-color: None")
            self.pushButtonDiff.hide()
        else:
            self.pushButtonDiff.show()
            self.pushButtonDiff.setEnabled(True)
            name = self.mdl.m_sca.getScalarDiffusivityName(self.scalar)
            exp = self.mdl.m_sca.getDiffFormula(self.scalar, self.zone_id)
            if exp:
                self.pushButtonDiff.setStyleSheet("background-color: green")
                self.pushButtonDiff.setToolTip(exp)
            else:
                self.pushButtonDiff.setStyleSheet("background-color: red")

        self.mdl.m_sca.setScalarDiffusivityChoice(self.scalar, choice)


    @pyqtSlot(str)
    def slotNameDiff(self, text):
        """
        Method to set the variance scalar choosed
        """
        choice = self.modelNameDiff.dicoV2M[str(text)]
        log.debug("slotStateDiff -> %s" % (text))
        self.scalar = str(text)
        self.lineEditDiff.setText(str(self.mdl.m_sca.getScalarDiffusivityInitialValue(self.scalar)))

        mdl = self.mdl.m_sca.getScalarDiffusivityChoice(self.scalar)

        self.modelDiff.setItem(str_model=mdl)

        if  mdl!= 'user_law':
            self.pushButtonDiff.hide()
            self.pushButtonDiff.setEnabled(False)
            self.pushButtonDiff.setStyleSheet("background-color: None")
        else:
            self.pushButtonDiff.show()
            self.pushButtonDiff.setEnabled(True)
            name = self.mdl.m_sca.getScalarDiffusivityName(self.scalar)
            exp = self.mdl.m_sca.getDiffFormula(self.scalar, self.zone_id)
            if exp:
                self.pushButtonDiff.setStyleSheet("background-color: green")
                self.pushButtonDiff.setToolTip(exp)
            else:
                self.pushButtonDiff.setStyleSheet("background-color: red")


    def __changeChoice(self, text, sym, tag):
        """
        Input variable state
        """
        __model  = getattr(self, "model"      + sym)
        __line   = getattr(self, "lineEdit"   + sym)
        __combo  = getattr(self, "comboBox"   + sym)
        __label  = getattr(self, "label"      + sym)
        __button = getattr(self, "pushButton" + sym)
        __labelu = getattr(self, "labelUnit"  + sym)
        __labelv = getattr(self, "labelVar"  + sym)

        choice = __model.dicoV2M[text]
        log.debug("__changeChoice -> %s, %s" % (text, choice))

        if choice != 'user_law':
            __button.hide()
            __button.setEnabled(False)
            __button.setStyleSheet("background-color: None")
        else:
            __button.setEnabled(True)
            __button.show()
            exp = None
            if sym == "Rho":
                exp = self.mdl.getFormula('density', self.zone_id)
            elif sym == "Mu":
                exp = self.mdl.getFormula('molecular_viscosity', self.zone_id)
            elif sym == "Cp":
                exp = self.mdl.getFormula('specific_heat', self.zone_id)
            elif sym == "Viscv0":
                exp = self.mdl.getFormula('volume_viscosity', self.zone_id)
            elif sym == "Al":
                exp = self.mdl.getFormula('thermal_conductivity', self.zone_id)
            elif sym == "Diff":
                name = self.mdl.m_sca.getScalarDiffusivityName(self.scalar)
                exp = self.mdl.m_sca.getDiffFormula(self.scalar, self.zone_id)

            if exp:
                __button.setStyleSheet("background-color: green")
                __button.setToolTip(exp)
            else:
                __button.setStyleSheet("background-color: red")
        if choice == 'thermal_law':
            __line.hide()
            __label.hide()
            __labelu.hide()
            __labelv.hide()
        else:
            __line.show()
            __label.show()
            __labelu.show()
            __labelv.show()

        if self.is_main_zone:
            self.mdl.setPropertyMode(tag, choice)


    @pyqtSlot(str)
    def slotRho(self, text):
        """
        Update the density
        """
        if self.lineEditRho.validator().state == QValidator.Acceptable:
            rho = from_qvariant(text, float)
            self.mdl.setInitialValueDensity(rho)

    @pyqtSlot(str)
    def slotRho1(self, text):
        """
        Update the density of fluid 1 for VoF module
        """
        if self.lineEditRho1.validator().state == QValidator.Acceptable:
            rho = from_qvariant(text, float)
            self.mdl.setVofValueDensity(0, rho)

    @pyqtSlot(str)
    def slotRho2(self, text):
        """
        Update the density of fluid 2 for VoF module
        """
        if self.lineEditRho2.validator().state == QValidator.Acceptable:
            rho = from_qvariant(text, float)
            self.mdl.setVofValueDensity(1, rho)

    @pyqtSlot(str)
    def slotMu(self, text):
        """
        Update the molecular viscosity
        """
        if self.lineEditMu.validator().state == QValidator.Acceptable:
            mu = from_qvariant(text, float)
            self.mdl.setInitialValueViscosity(mu)


    @pyqtSlot(str)
    def slotMu1(self, text):
        """
        Update the molecular viscosity
        """
        if self.lineEditMu1.validator().state == QValidator.Acceptable:
            mu = from_qvariant(text, float)
            self.mdl.setVofValueViscosity(0, mu)


    @pyqtSlot(str)
    def slotMu2(self, text):
        """
        Update the molecular viscosity
        """
        if self.lineEditMu2.validator().state == QValidator.Acceptable:
            mu = from_qvariant(text, float)
            self.mdl.setVofValueViscosity(1, mu)


    @pyqtSlot(str)
    def slotCp(self, text):
        """
        Update the specific heat
        """
        if self.lineEditCp.validator().state == QValidator.Acceptable:
            cp = from_qvariant(text, float)
            self.mdl.setInitialValueHeat(cp)


    @pyqtSlot(str)
    def slotViscv0(self, text):
        """
        Update the volumic viscosity
        """
        if self.lineEditViscv0.validator().state == QValidator.Acceptable:
            viscv0 = from_qvariant(text, float)
            self.mdl.setInitialValueVolumeViscosity(viscv0)


    @pyqtSlot(str)
    def slotAl(self, text):
        """
        Update the thermal conductivity
        """
        if self.lineEditAl.validator().state == QValidator.Acceptable:
            al = from_qvariant(text, float)
            self.mdl.setInitialValueCond(al)


    @pyqtSlot(str)
    def slotDiftl0(self, text):
        """
        Update the thermal conductivity
        """
        if self.lineEditDiftl0.validator().state == QValidator.Acceptable:
            diftl0 = from_qvariant(text, float)
            self.mdl.setInitialValueDyn(diftl0)


    @pyqtSlot(str)
    def slotDiff(self, text):
        """
        Update the thermal conductivity
        """
        if self.lineEditDiff.validator().state == QValidator.Acceptable:
            diff = from_qvariant(text, float)
            self.mdl.m_sca.setScalarDiffusivityInitialValue(self.scalar, diff)


    @pyqtSlot()
    def slotFormulaRho(self):
        """
        User formula for density
        """
        exp, req, sca, symbols_rho = \
            self.mdl.getFormulaRhoComponents(self.zone_id)

        self.m_th = ThermalScalarModel(self.case)
        s = self.m_th.getThermalScalarName()
        mdl = self.m_th.getThermalScalarModel()
        if mdl == "off":
            exa = FluidCharacteristicsView.density_wo
        elif mdl == "temperature_celsius":
            TempInContext = "("+s+" + 273.15)"
            exa = FluidCharacteristicsView.density.replace("temperature", TempInContext)
        elif mdl == "enthalpy":
            exa = FluidCharacteristicsView.density_h
        else:
            exa = FluidCharacteristicsView.density

        dialog = QMegEditorView(parent        = self,
                                function_type = 'vol',
                                zone_name     = self.zone_name,
                                variable_name = 'density',
                                expression    = exp,
                                required      = req,
                                symbols       = symbols_rho,
                                known_fields  = sca,
                                examples      = exa)

        if dialog.exec_():
            result = dialog.get_result()
            log.debug("slotFormulaRho -> %s" % str(result))
            self.mdl.setFormula('density', str(result), self.zone_id)
            self.pushButtonRho.setToolTip(result)
            self.pushButtonRho.setStyleSheet("background-color: green")


    @pyqtSlot()
    def slotFormulaMu(self):
        """
        User formula for molecular viscosity
        """

        exp, req, sca, symbols_mu = \
                self.mdl.getFormulaMuComponents(self.zone_id)

        self.m_th = ThermalScalarModel(self.case)
        s = self.m_th.getThermalScalarName()
        mdl = self.m_th.getThermalScalarModel()
        if mdl == "off":
            exa = FluidCharacteristicsView.molecular_viscosity_wo
        elif mdl == "temperature_celsius":
            TempInContext = "("+s+" + 273.15)"
            exa = FluidCharacteristicsView.molecular_viscosity.replace("temperature", TempInContext)
        elif mdl == "enthalpy":
            exa = FluidCharacteristicsView.molecular_viscosity_h
        else:
            exa = FluidCharacteristicsView.molecular_viscosity

        dialog = QMegEditorView(parent        = self,
                                function_type = 'vol',
                                zone_name     = self.zone_name,
                                variable_name = 'molecular_viscosity',
                                expression    = exp,
                                required      = req,
                                symbols       = symbols_mu,
                                known_fields  = sca,
                                examples      = exa)

        if dialog.exec_():
            result = dialog.get_result()
            log.debug("slotFormulaMu -> %s" % str(result))
            self.mdl.setFormula('molecular_viscosity', str(result), self.zone_id)
            self.pushButtonMu.setToolTip(result)
            self.pushButtonMu.setStyleSheet("background-color: green")


    @pyqtSlot()
    def slotFormulaCp(self):
        """
        User formula for specific heat
        """
        exp, req, sca, symbols_cp = \
                self.mdl.getFormulaCpComponents(self.zone_id)

        exa = FluidCharacteristicsView.specific_heat

        dialog = QMegEditorView(parent        = self,
                                function_type = 'vol',
                                zone_name     = self.zone_name,
                                variable_name = 'specific_heat',
                                expression    = exp,
                                required      = req,
                                symbols       = symbols_cp,
                                known_fields  = sca,
                                examples      = exa)

        if dialog.exec_():
            result = dialog.get_result()
            log.debug("slotFormulaRho -> %s" % str(result))
            self.mdl.setFormula('specific_heat', str(result), self.zone_id)
            self.pushButtonCp.setToolTip(result)
            self.pushButtonCp.setStyleSheet("background-color: green")


    @pyqtSlot()
    def slotFormulaViscv0(self):
        """
        User formula for volumic viscosity
        """
        exp, req, sca, symbols_viscv0 = \
                self.mdl.getFormulaViscv0Components(self.zone_id)

        exa = FluidCharacteristicsView.volume_viscosity

        dialog = QMegEditorView(parent        = self,
                                function_type = 'vol',
                                zone_name     = self.zone_name,
                                variable_name = 'volume_viscosity',
                                expression    = exp,
                                required      = req,
                                symbols       = symbols_viscv0,
                                known_fields  = sca,
                                examples      = exa)

        if dialog.exec_():
            result = dialog.get_result()
            log.debug("slotFormulaViscv0 -> %s" % str(result))
            self.mdl.setFormula('volume_viscosity', str(result), self.zone_id)
            self.pushButtonViscv0.setToolTip(result)
            self.pushButtonViscv0.setStyleSheet("background-color: green")


    @pyqtSlot()
    def slotFormulaAl(self):
        """
        User formula for thermal conductivity
        """
        exp, req, sca, symbols_al = \
                self.mdl.getFormulaAlComponents(self.zone_id)

        self.m_th = ThermalScalarModel(self.case)
        s = self.m_th.getThermalScalarName()
        mdl = self.m_th.getThermalScalarModel()
        if mdl == "temperature_celsius":
            TempInContext = "("+s+" + 273.15)"
            exa = FluidCharacteristicsView.thermal_conductivity.replace("temperature", TempInContext)
        elif mdl == "enthalpy":
            exa = FluidCharacteristicsView.thermal_conductivity_h
        else:
            exa = FluidCharacteristicsView.thermal_conductivity

        dialog = QMegEditorView(parent        = self,
                                function_type = 'vol',
                                zone_name     = self.zone_name,
                                variable_name = 'thermal_conductivity',
                                expression    = exp,
                                required      = req,
                                symbols       = symbols_al,
                                known_fields  = sca,
                                examples      = exa)

        if dialog.exec_():
            result = dialog.get_result()
            log.debug("slotFormulaAl -> %s" % str(result))
            self.mdl.setFormula('thermal_conductivity', str(result), self.zone_id)
            self.pushButtonAl.setToolTip(result)
            self.pushButtonAl.setStyleSheet("background-color: green")


    @pyqtSlot()
    def slotFormulaDiff(self):
        """
        User formula for the diffusion coefficient
        """
        exp, req, sca, sym = \
                self.mdl.getFormulaDiffComponents(self.scalar, self.zone_id)

        exa = ''

        dname = self.mdl.m_sca.getScalarDiffusivityName(self.scalar)

        dialog = QMegEditorView(parent        = self,
                                function_type = 'vol',
                                zone_name     = self.zone_name,
                                variable_name = dname,
                                expression    = exp,
                                required      = req,
                                symbols       = sym,
                                known_fields  = sca,
                                examples      = exa)

        if dialog.exec_():
            result = dialog.get_result()
            log.debug("slotFormulaDiff -> %s" % str(result))
            self.mdl.m_sca.setDiffFormula(self.scalar, str(result), self.zone_id)
            self.pushButtonDiff.setToolTip(result)
            self.pushButtonDiff.setStyleSheet("background-color: green")


#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------



# -*- coding: utf-8 -*-

#-------------------------------------------------------------------------------

# This file is part of Code_Saturne, a general-purpose CFD tool.
#
# Copyright (C) 1998-2021 EDF S.A.
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; either version 2 of the License, or (at your option) any later
# version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.

#-------------------------------------------------------------------------------

"""
This module manages plots :

This module defines the following classes:
- ManagePlotterModel
"""

#-------------------------------------------------------------------------------
# Library modules import
#-------------------------------------------------------------------------------

import sys, unittest

#-------------------------------------------------------------------------------
# Application modules import
#-------------------------------------------------------------------------------

from code_saturne.model.XMLvariables import Model, Variables
from code_saturne.model.XMLmodel import ModelTest

#-------------------------------------------------------------------------------
# Model class
#-------------------------------------------------------------------------------

class ManagePlotterModel(Model):
    """
    Class for Variables and Scalar model initialization.
    """
    def __init__(self, case):
        """
        Constructor
        """
        self.case = case
        self.repo = self.case.xmlGetNode("studymanager").xmlGetString('repository')


    def defaultInitialValues(self):
        """
        Return in a dictionnary which contains default values.
        """
        default = {}
        default['subplot_name']      = ""
        default['subplot_legstatus'] = "on"
        default['figure_name']       = "Figure"
        default['figure_title']      = ""
        default['figure_fmt']        = "pdf"
        default['color']             = "bedf"

        return default


    def getStudyList(self):
        """
        Return list of studies in xml file
        """
        list_study = self.case.xmlGetNodeList("study")
        lst = []
        for node in list_study:
            lst.append(node['label'])
        return lst


    def getSubplotList(self, study):
        """
        Return list of subplots for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = []
        for node in study_node.xmlGetNodeList("subplot"):
            lst.append(node['id'])
        return lst


    def getFigureList(self, study):
        """
        Return list of figures for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = []
        for node in study_node.xmlGetNodeList("figure"):
            lst.append(node['id'])
        return lst


    def getMeasurementList(self, study):
        """
        Return list of measurements for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = []
        for node in study_node.xmlGetNodeList("measurement"):
            lst.append((node['file'], node['path']))
        return lst


    def getCaseList(self, study):
        """
        Return list of case for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = []
        for nn in study_node.xmlGetNodeList("case"):
            lst.append(int(nn['id']))
        return lst


    def addSubplot(self, study):
        """
        add a subplot for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        idx = len(study_node.xmlGetNodeList("subplot"))
        study_node.xmlInitChildNode("subplot", id = idx)
        return idx


    def delSubplot(self, study, idx):
        """
        suppress a subplot for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        node = study_node.xmlGetNode("subplot", id = idx)
        node.xmlRemoveNode()
        for nn in study_node.xmlGetNodeList("subplot"):
            try:
                if int(nn['id']) > idx:
                    nn['id'] = str(int(nn['id']) - 1)
            except:
                pass

        for nn in study_node.xmlGetNodeList("figure"):
            lst = nn['idlist']
            new_lst = ''
            if lst:
                for idl in lst:
                    if idl != " ":
                        new_idl = idl
                        if int(idl) > idx:
                            new_idl = str(int(idl) - 1)
                        if new_lst != "":
                            new_lst = new_lst + " " + str(new_idl)
                        else:
                            new_lst = str(new_idl)
                    nn['idlist'] = new_lst

        for nn in study_node.xmlGetNodeList("plot"):
            lst = nn['spids']
            new_lst = ''
            if lst:
                for idl in lst:
                    if idl != " ":
                        new_idl = idl
                        if int(idl) > idx:
                            new_idl = str(int(idl) - 1)
                        if new_lst != "":
                            new_lst = new_lst + " " + str(new_idl)
                        else:
                            new_lst = str(new_idl)
                    nn['spids'] = new_lst


    def getSubplotTitle(self, study, idx):
        """
        Return title for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        name = subplot['title']
        if not name:
            name = self.defaultInitialValues()['subplot_name']
        return name


    def setSubplotTitle(self, study, idx, title):
        """
        Set title for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        if title != "":
            subplot['title'] = title


    def getSubplotXLabel(self, study, idx):
        """
        Return X label for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        name = subplot['xlabel']
        if not name:
            name = ""
        return name


    def setSubplotXLabel(self, study, idx, title):
        """
        Set X label for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        subplot['xlabel'] = title


    def getSubplotYLabel(self, study, idx):
        """
        Return Y label for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        name = subplot['ylabel']
        if not name:
            name = ""
        return name


    def setSubplotYLabel(self, study, idx, title):
        """
        Set Y label for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        subplot['ylabel'] = title


    def getSubplotLegPos(self, study, idx):
        """
        Return legend position for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        name = subplot['legpos']
        if not name:
            name = ""
        return name


    def setSubplotLegPos(self, study, idx, title):
        """
        Set legend position for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        subplot['legpos'] = title


    def getSubplotLegStatus(self, study, idx):
        """
        Return legend status for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        status = subplot['legstatus']
        if not status:
            status = self.defaultInitialValues()['subplot_legstatus']
            self.setSubplotLegStatus(study, idx, status)
        return status


    def setSubplotLegStatus(self, study, idx, status):
        """
        Set legend status for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        subplot['legstatus'] = status


    def getSubplotXLim(self, study, idx):
        """
        Return X limit for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        status = subplot['xlim']
        if not status:
            status = ""
        return status


    def setSubplotXLim(self, study, idx, pos):
        """
        Set X limit for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        if pos != "":
            subplot['xlim'] = pos


    def getSubplotYLim(self, study, idx):
        """
        Return Y limit for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        status = subplot['ylim']
        if not status:
            status = ""
        return status


    def setSubplotYLim(self, study, idx, pos):
        """
        Set Y limit for a subplot
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("subplot", id = idx)
        if pos != "":
            subplot['ylim'] = pos


#-------------------------------------------------------------------------------
# Figure
#-------------------------------------------------------------------------------
    def addFigure(self, study):
        """
        add a subplot for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        idx = len(study_node.xmlGetNodeList("figure"))
        study_node.xmlInitChildNode("figure", id = idx)
        return idx


    def delFigure(self, study, idx):
        """
        suppress a subplot for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        node = study_node.xmlGetNode("figure", id = idx)
        node.xmlRemoveNode()
        for nn in study_node.xmlGetNodeList("figure"):
            try:
                if int(nn['id']) > idx:
                    nn['id'] = str(int(nn['id']) - 1)
            except:
                pass


    def getFigureName(self, study, idx):
        """
        Return name for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        name = subplot['name']
        if not name:
            name = self.defaultInitialValues()['figure_name']
            self.setFigureName(study, idx, name)
        return name


    def setFigureName(self, study, idx, name):
        """
        Set name for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        subplot['name'] = name


    def getFigureTitle(self, study, idx):
        """
        Return title for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        name = subplot['title']
        if not name:
            name = self.defaultInitialValues()['figure_title']
            self.setFigureTitle(study, idx, name)
        return name


    def setFigureTitle(self, study, idx, name):
        """
        Set title for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        if name != "":
            subplot['title'] = name
        elif subplot['title']:
            subplot['title'] = name


    def getFigureRow(self, study, idx):
        """
        Return number of row for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        row = subplot['nbrow']
        if not row:
            row = ""
        return row


    def setFigureRow(self, study, idx, row):
        """
        Set number of row for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        if row != "":
            subplot['nbrow'] = row


    def getFigureColumn(self, study, idx):
        """
        Return number of column for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        col = subplot['nbcol']
        if not col:
            col = ""
        return col


    def setFigureColumn(self, study, idx, col):
        """
        Set number of column for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        if col != "":
            subplot['nbcol'] = col


    def getFigureFormat(self, study, idx):
        """
        Return format for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        name = subplot['format']
        if not name:
            name = self.defaultInitialValues()['figure_fmt']
            self.setFigureFormat(study, idx, name)
        return name


    def setFigureFormat(self, study, idx, fmt):
        """
        Set format for a figure
        """
        self.isInList(fmt, ("png", "pdf"))
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        subplot['format'] = fmt


    def getFigureIdList(self, study, idx):
        """
        Return subplot id list for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        lst = subplot['idlist']
        if not lst:
            lst = ""
        return lst


    def setFigureIdList(self, study, idx, idlist):
        """
        Set subplot id list for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        subplot = study_node.xmlGetNode("figure", id = idx)
        subplot['idlist'] = idlist


#-------------------------------------------------------------------------------
# Measurement
#-------------------------------------------------------------------------------
    def addMeasurementFile(self, study, fle):
        """
        add a measurement file for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        study_node.xmlInitChildNode("measurement", file=fle)


    def delMeasurementFile(self, study, idx):
        """
        suppress a measurement file for a study
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[idx]
        node.xmlRemoveNode()


    def addMeasurementPlot(self, study, measurement_idx):
        """
        add a plot for a measurement file
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        lst2 = study_node.xmlGetNodeList("plot")
        idx = len(lst2)
        node.xmlInitChildNode("plot", id = idx)
        return idx


    def deleteMeasurementPlot(self, study, measurement_idx, idx):
        """
        suppress a plot for a measurement file
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        nn = node.xmlGetNode("plot", id = idx)
        nn.xmlRemoveNode()

        idx = 0
        for n in study_node.xmlGetNodeList("plot"):
            n['id'] = idx
            idx = idx + 1


    def getMeasurementPlotList(self, study, name, path):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        measurement_node = study_node.xmlGetNode("measurement", file=name, path=path)
        lst = []
        for node in measurement_node.xmlGetNodeList("plot"):
            lst.append(node['id'])
        return lst


    def getMeasurementColor(self, study, measurement_idx, idx):
        """
        Return plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        color = plot['color']
        if not color:
            color = self.defaultInitialValues()['color']
            self.setMeasurementColor(study, measurement_idx, idx, color)
        return color


    def setMeasurementColor(self, study, measurement_idx, idx, color):
        """
        Set plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['color'] = color


    def getMeasurementFormat(self, study, measurement_idx, idx):
        """
        Return plot format
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        fmt = plot['fmt']
        if not fmt:
            fmt = ""
        return fmt


    def setMeasurementFormat(self, study, measurement_idx, idx, fmt):
        """
        Set plot format
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['fmt'] = fmt


    def getMeasurementLegend(self, study, measurement_idx, idx):
        """
        Return plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        legend = plot['legend']
        if not legend:
            legend = ""
        return legend


    def setMeasurementLegend(self, study, measurement_idx, idx, legend):
        """
        Set plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['legend'] = legend


    def getMeasurementXcol(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['xcol']
        if not val:
            val = ""
        return val


    def setMeasurementXcol(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['xcol'] = val


    def getMeasurementYcol(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['ycol']
        if not val:
            val = ""
        return val


    def setMeasurementYcol(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['ycol'] = val


    def getMeasurementWidth(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['linewidth']
        if not val:
            val = ""
        return val


    def setMeasurementWidth(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['linewidth'] = val


    def getMeasurementMarker(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['markersize']
        if not val:
            val = ""
        return val


    def setMeasurementMarker(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['markersize'] = val


    def getMeasurementXerr(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['xerr']
        if not val:
            val = ""
        return val


    def setMeasurementXerr(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['xerr'] = val


    def getMeasurementXerrp(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['xerrp']
        if not val:
            val = ""
        return val


    def setMeasurementXerrp(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['xerrp'] = val


    def getMeasurementYerr(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['yerr']
        if not val:
            val = ""
        return val


    def setMeasurementYerr(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['yerr'] = val


    def getMeasurementYerrp(self, study, measurement_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        val = plot['yerrp']
        if not val:
            val = ""
        return val


    def setMeasurementYerrp(self, study, measurement_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['yerrp'] = val


    def getMeasurementIdList(self, study, measurement_idx, idx):
        """
        Return subplot id list for a plot of a measurement
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        lst = plot['spids']
        if not lst:
            lst = ""
        return lst


    def setMeasurementIdList(self, study, measurement_idx, idx, idlist):
        """
        Set subplot id list for a plot of a measurement
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("measurement")
        node = lst[measurement_idx]
        plot = node.xmlGetNode("plot", id = idx)
        plot['spids'] = idlist


#-------------------------------------------------------------------------------
# Case
#-------------------------------------------------------------------------------
    def getCaseName(self, study, idx):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        nn = study_node.xmlGetNode("case", id = idx)
        return nn['label']


    def getCaseDataList(self, study, idx):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        nn = study_node.xmlGetNode("case", id = idx)
        lst = []
        for n in nn.xmlGetNodeList("data"):
            lst.append(n['file'])
        return lst


    def addCaseDataFile(self, study, case_idx, name):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        nn = study_node.xmlGetNode("case", id = case_idx)
        nn.xmlInitChildNode("data", dest="", file=name)


    def delCaseDataFile(self, study, case_idx, idx):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        nn = study_node.xmlGetNode("case", id = case_idx)
        lst = nn.xmlGetNodeList("data")
        node = lst[idx]
        node.xmlRemoveNode()


    def addAssociatedCasePlot(self, study, case_idx, data_idx):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        nn = study_node.xmlGetNode("case", id = case_idx)
        node = nn.xmlGetNodeList("data")[data_idx]
        idx = len(study_node.xmlGetNodeList("plot"))
        node.xmlInitChildNode("plot", id = idx)
        return idx


    def delAssociatedCasePlot(self, study, case_idx, data_idx, idx):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        nn = study_node.xmlGetNode("case", id = case_idx)
        node = nn.xmlGetNodeList("data")[data_idx]
        lst = node.xmlGetNodeList("plot")
        n = lst[idx]
        n.xmlRemoveNode()

        idx = 0
        for n in study_node.xmlGetNodeList("plot"):
            n['id'] = idx
            idx = idx + 1


    def getCasePlotList(self, study, idx, data_idx):
        """
        """
        study_node = self.case.xmlGetNode("study", label = study)
        case = study_node.xmlGetNode("case", id = idx)
        node = case.xmlGetNodeList("data")[data_idx]
        lst = []
        for n in node.xmlGetNodeList("plot"):
            lst.append(n['id'])
        return lst


    def getAssociatedCaseColor(self, study, case_idx, data_idx, idx):
        """
        Return plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        color = plot['color']
        if not color:
            color = self.defaultInitialValues()['color']
            self.setAssociatedCaseColor(study, case_idx, data_idx, idx, color)
        return color


    def setAssociatedCaseColor(self, study, case_idx, data_idx, idx, color):
        """
        Set plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['color'] = color


    def getAssociatedCaseFormat(self, study, case_idx, data_idx, idx):
        """
        Return plot format
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        fmt = plot['fmt']
        if not fmt:
            fmt = ""
        return fmt


    def setAssociatedCaseFormat(self, study, case_idx, data_idx, idx, fmt):
        """
        Set plot format
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['fmt'] = fmt


    def getAssociatedCaseLegend(self, study, case_idx, data_idx, idx):
        """
        Return plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        legend = plot['legend']
        if not legend:
            legend = ""
        return legend


    def setAssociatedCaseLegend(self, study, case_idx, data_idx, idx, legend):
        """
        Set plot color
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['legend'] = legend


    def getAssociatedCaseXcol(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['xcol']
        if not val:
            val = ""
        return val


    def setAssociatedCaseXcol(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['xcol'] = val


    def getAssociatedCaseYcol(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['ycol']
        if not val:
            val = ""
        return val


    def setAssociatedCaseYcol(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['ycol'] = val


    def getAssociatedCaseWidth(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['linewidth']
        if not val:
            val = ""
        return val


    def setAssociatedCaseWidth(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['linewidth'] = val


    def getAssociatedCaseMarker(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['markersize']
        if not val:
            val = ""
        return val


    def setAssociatedCaseMarker(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['markersize'] = val


    def getAssociatedCaseXerr(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['xerr']
        if not val:
            val = ""
        return val


    def setAssociatedCaseXerr(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['xerr'] = val


    def getAssociatedCaseXerrp(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['xerrp']
        if not val:
            val = ""
        return val


    def setAssociatedCaseXerrp(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['xerrp'] = val


    def getAssociatedCaseYerr(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['yerr']
        if not val:
            val = ""
        return val


    def setAssociatedCaseYerr(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['yerr'] = val


    def getAssociatedCaseYerrp(self, study, case_idx, data_idx, idx):
        """
        Return plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        val = plot['yerrp']
        if not val:
            val = ""
        return val


    def setAssociatedCaseYerrp(self, study, case_idx, data_idx, idx, val):
        """
        Set plot val
        """
        study_node = self.case.xmlGetNode("study", label = study)
        lst = study_node.xmlGetNodeList("case")
        node = lst[case_idx]
        n = node.xmlGetNodeList("data")[data_idx]
        plot = n.xmlGetNode("plot", id = idx)
        plot['yerrp'] = val


    def getCasePlotId(self, study, case_idx, data_idx, idx):
        """
        Return subplot id list for a case
        """
        study_node = self.case.xmlGetNode("study", label = study)
        case = study_node.xmlGetNode("case", id = case_idx)
        node = case.xmlGetNodeList("data")[data_idx]
        n = node.xmlGetNodeList("plot")[idx]
        return n['id']


    def getCaseIdList(self, study, case_idx, data_idx, plot_id):
        """
        Return subplot id list for a case
        """
        study_node = self.case.xmlGetNode("study", label = study)
        case = study_node.xmlGetNode("case", id = case_idx)
        node = case.xmlGetNodeList("data")[data_idx]
        n = node.xmlGetNode("plot", id = plot_id)
        lst = n['spids']
        if not lst:
            lst = ""
        return lst


    def setCaseIdList(self, study, case_idx, data_idx, plot_id, idlist):
        """
        Set subplot id list for a figure
        """
        study_node = self.case.xmlGetNode("study", label = study)
        case = study_node.xmlGetNode("case", id = case_idx)
        node = case.xmlGetNodeList("data")[data_idx]
        plot = node.xmlGetNode("plot", id = plot_id)
        plot['spids'] = idlist

#-------------------------------------------------------------------------------
# End
#-------------------------------------------------------------------------------
